import pandas as pd
import numpy as np
import pytest

from pandas.testing import assert_series_equal
from numpy.testing import assert_equal
from siuba.tests.helpers import assert_equal_query

from siuba.siu import _
from siuba.dply.verbs import case_when, mutate


DATA = pd.DataFrame({
    'x': [0,1,2],
    'y': [10, 11, 12]
    })


@pytest.fixture
def data():
    return DATA.copy()


@pytest.mark.parametrize("k,v, dst", [
    (True, 1,     [1]*3),
    (True, False, [False]*3),
    (True, _.y,   [10, 11, 12]),
    (True, lambda _: _.y, [10, 11, 12]),
    (_.x < 2, 0,  [0, 0, None]),
    (_.x < 2, "small", ["small", "small", None]),
    (_.x < 2, _.y, [10, 11, None]),
    (lambda _: _.x < 2, 0,  [0, 0, None]),
    #(np.array([True, True, False]), 0, [0, 0, None])
    ])
def test_case_when_single_cond(backend, data, k, v, dst):
    src = backend.load_df(data)
    query = mutate(_, res = case_when(_, {k: v}))

    assert_equal_query(src, query, data.assign(res = dst))


def test_case_when_multiple_clauses(backend, data):
    src = backend.load_df(data)
    query = mutate(_, res = case_when({_.x == 0: "zero", _.x > 1: "big", True: "small"}))

    assert_equal_query(src, query, data.assign(res = ["zero", "small", "big"]))


def test_case_when_cond_order(data):
    out = case_when(data, {
        lambda _: _.x < 2  :  0,
        True               : 999
        })

    assert_series_equal(out, pd.Series([0, 0, 999]))




"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/dbplyr/blob/master/tests/testthat/test-verb-mutate.R
"""
    
from siuba import _, mutate, select, group_by, summarize, filter, arrange
from siuba.dply.vector import row_number

import pytest
from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql
from string import ascii_lowercase 

DATA = data_frame(a = [1,2,3], b = [9,8,7])

@pytest.fixture(scope = "module")
def dfs(backend):
    return backend.load_df(DATA)

@pytest.mark.parametrize("query, output", [
    (mutate(x = _.a + _.b), DATA.assign(x = [10, 10, 10])),
    (mutate(x = _.a + _.b) >> summarize(ttl = _.x.sum().astype(float)), data_frame(ttl = 30.0)),
    (mutate(x = _.a + 1, y = _.b - 1), DATA.assign(x = [2,3,4], y = [8,7,6])),
    (mutate(x = _.a + 1) >> mutate(y = _.b - 1), DATA.assign(x = [2,3,4], y = [8,7,6])),
    (mutate(x = _.a + 1, y = _.x + 1), DATA.assign(x = [2,3,4], y = [3,4,5])),
    (mutate(_, _.a), DATA.copy()),
    (mutate(_, _.a, _.a), DATA.copy()),
    ])
def test_mutate_basic(dfs, query, output):
    assert_equal_query(dfs, query, output)

@pytest.mark.parametrize("query, output", [
    (mutate(x = 1), DATA.assign(x = 1)),
    (mutate(x = "a"), DATA.assign(x = "a")),
    (mutate(x = 1.2), DATA.assign(x = 1.2)),
    ])
def test_mutate_literal(dfs, query, output):
    assert_equal_query(dfs, query, output)


def test_select_mutate_filter(dfs):
    assert_equal_query(
            dfs,
            select(_.x == _.a) >> mutate(y = _.x * 2) >> filter(_.y == 2),
            data_frame(x = 1, y = 2)
            )

@backend_sql
def test_mutate_smart_nesting(backend, dfs):
    # y and z both use x, so should create only 1 extra query
    lazy_tbl = dfs >> mutate(x = _.a + 1, y = _.x + 1, z = _.x + 1)

    # should have form
    # SELECT ..., x + 1 as y, x + 1 as z FROM (
    # SELECT ..., a + 1 as x FROM
    # <TABLENAME>) some_alias
    inner_alias = lazy_tbl.last_op.froms[0]
    inner_select = inner_alias.element
    orig_table = inner_select.froms[0]

    assert orig_table is lazy_tbl.tbl


def test_mutate_reassign_column_ordering(dfs):
    assert_equal_query(
            dfs,
            mutate(c = 3, a = 1),
            data_frame(a = [1,1,1], b = [9,8,7], c = [3,3,3])
            )

def test_mutate_reassign_all_cols_keeps_rowsize(dfs):
    assert_equal_query(
            dfs,
            mutate(a = 1, b = 2),
            data_frame(a = [1,1,1], b = [2,2,2])
            )


def test_mutate_grouped_pandas_no_dropna():
    src = data_frame(x = [1, 2], g = [None, None])

    assert_equal_query(
        src,
        group_by(_.g) >> mutate(res = _.x + 1),
        data_frame(x = [1, 2], g = [None, None], res = [2, 3])
    )


@backend_sql
def test_mutate_window_funcs(backend):
    data = data_frame(idx = range(0, 4), x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)
    assert_equal_query(
            dfs,
            arrange(_.idx) >> group_by(_.g) >> mutate(row_num = row_number(_).astype(float)),
            data.assign(row_num = [1., 2, 1, 2])
            )


def test_mutate_using_agg_expr(backend):
    data = data_frame(x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)
    assert_equal_query(
            dfs,
            group_by(_.g) >> mutate(y = _.x - _.x.mean()),
            data.assign(y = [-.5, .5, -.5, .5])
            )

@backend_sql # TODO: pandas outputs a int column
def test_mutate_using_cuml_agg(backend):
    data = data_frame(idx = range(0, 4), x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)

    # cuml window without arrange before generates warning
    with pytest.warns(None):
        assert_equal_query(
                dfs,
                arrange(_.idx) >> group_by(_.g) >> mutate(y = _.x.cumsum()),
                data.assign(y = [1, 3, 3, 7]),
                check_dtype=False       # bigquery returns int, postgres float
                )

def test_mutate_overwrites_prev(backend):
    # TODO: check that query doesn't generate a CTE
    dfs = backend.load_df(data_frame(x = range(1, 5), g = [1,1,2,2]))
    assert_equal_query(
            dfs,
            mutate(x = _.x + 1) >> mutate(x = _.x + 1),
            data_frame(x = [3,4,5,6], g = [1,1,2,2])
            )


def test_mutate_after_summarize_on_non_derived_column(backend):
    dfs = backend.load_df(data_frame(x = range(1, 5), g = [1,2,2,2]))

    query = group_by(_.g) >> summarize(avg = _.x.min()) >> mutate(avg_g = _.g.mean())
    assert_equal_query(
            dfs,
            query,
            data_frame(g = [1,2], avg = [1,2], avg_g = 1.5)
            )
    

def test_mutate_after_summarize_on_derived_column(backend):
    dfs = backend.load_df(data_frame(x = range(1, 5), g = [1,2,2,2]))

    query = group_by(_.g) >> summarize(avg = _.x.min()) >> mutate(avg_avg = _.avg.mean())
    assert_equal_query(
            dfs,
            query,
            data_frame(g = [1,2], avg = [1,2], avg_avg = 1.5)
            )


def test_mutate_after_summarize_limits_column_access(backend):
    dfs = backend.load_df(data_frame(x = range(1, 5), g = [1,2,2,2]))

    query = group_by(_.g) >> summarize(avg = _.x.min()) >> mutate(x2 = _.x + 1)

    with pytest.raises(AttributeError) as exc_info:
        query(dfs)

    
    assert "x" in exc_info.value.args[0]



from siuba.dply.verbs import filter

from ..backend import LazyTbl
from ..translate import ColumnCollection
from ..utils import _sql_select

from sqlalchemy import sql
from siuba.siu import Call

from siuba.dply.across import _set_data_context


@filter.register(LazyTbl)
def _filter(__data, *args):
    # Note: currently always produces 2 additional select statements,
    #       1 for window/aggs, and 1 for the where clause

    sel = __data.last_op.alias()                   # original select
    win_sel = sel.select()

    conds = []
    windows = []
    with _set_data_context(__data, window=True):
        for ii, arg in enumerate(args):

            if isinstance(arg, Call):
                new_call = __data.shape_call(arg, verb_name = "Filter", arg_name = ii)
                #var_cols = new_call.op_vars(attr_calls = False)

                # note that a new win_sel is returned, w/ window columns appended
                col_expr, win_cols, win_sel = __data.track_call_windows(
                        new_call,
                        sel.columns,
                        window_cte = win_sel
                        )

                if isinstance(col_expr, ColumnCollection):
                    conds.extend(col_expr)
                else:
                    conds.append(col_expr)

                windows.extend(win_cols)
                
            else:
                conds.append(arg)

    bool_clause = sql.and_(*conds)

    # first cte, windows ----
    if len(windows):
        
        win_alias = win_sel.alias()

        # move non-window functions to refer to win_sel clause (not the innermost) ---
        bool_clause = sql.util.ClauseAdapter(win_alias) \
                .traverse(bool_clause)

        orig_cols = [win_alias.columns[k] for k in sel.columns.keys()]
    else:
        orig_cols = [sel]
    
    # create second cte ----
    filt_sel = _sql_select(orig_cols).where(bool_clause)
    return __data.append_op(filt_sel)



import pandas as pd

from siuba.siu import Call,   MetaArg, BinaryOp
from collections import OrderedDict
from itertools import chain
from functools import singledispatch

from typing import List

class Var:
    def __init__(self, name: "str | int | slice | Call", negated = False, alias = None):
        if not isinstance(name, (str, int, slice, Call)):
            raise TypeError(f"Var name cannot be type: {type(name)}.")
        self.name = name
        self.negated = negated
        self.alias = alias

    def __neg__(self):
        return self.to_copy(negated = not self.negated)


    def __invert__(self):
        return self.to_copy(negated = not self.negated)

    def __eq__(self, x):
        name = x.name if isinstance(x, Var) else x
        return self.to_copy(name = name, negated = False, alias = self.name)

    def __call__(self, *args, **kwargs):
        call = Call('__call__',
                    BinaryOp('__getattr__', MetaArg("_"), self.name),
                    *args,
                    **kwargs
                    )

        return self.to_copy(name = call)


    def __repr__(self):
        cls_name = self.__class__.__name__
        sig = f"{repr(self.name)}, negated={self.negated}, alias={self.alias}"
        return f"{cls_name}({sig})"

    def __str__(self):
        op = "-" if self.negated else ""
        pref = self.alias + " = " if self.alias else ""
        return "{pref}{op}{self.name}".format(pref = pref, op = op, self = self)

    def to_copy(self, **kwargs):
        return self.__class__(**{**self.__dict__, **kwargs})


class VarAnd(Var):
    name: "tuple[Var]"

    def __init__(self, name: "tuple[Var]", negated=False, alias=None):
        self.name = name
        self.negated = negated

        bad_var = [x for x in name if not isinstance(x, Var)]

        if any(bad_var):
            raise TypeError(f"VarAnd expects a tuple of Var, but saw entries: {bad_var}")

        if alias is not None:
            raise TypeError("alias must be none for VarAnd (extended slice syntax)")

        self.alias = None

    def __eq__(self, x):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    def flatten(self) -> "tuple[Var]":
        res = []
        for var in self.name:
            neg_var = ~var if self.negated else var
            if isinstance(neg_var, VarAnd):
                res.extend(neg_var.flatten())
            else:
                res.append(neg_var)

        return tuple(res)



class VarList:
    def __getattr__(self, x):
        return Var(x)

    def __getitem__(self, x):
        if not isinstance(x, tuple):
            return Var(x) if not isinstance(x, Var) else x
        else:
            res = [el if isinstance(el, Var) else Var(el) for el in x]
            return VarAnd(tuple(res))


def var_slice(colnames, x):
    """Return indices in colnames correspnding to start and stop of slice."""
    # TODO: produces bahavior similar to df.loc[:, "V1":"V3"], but can reverse
    # TODO: make DRY
    # TODO: reverse not including end points
    if isinstance(x.start, Var):
        start_indx = (colnames == x.start.name).idxmax()
    elif isinstance(x.start, str):
        start_indx = (colnames == x.start).idxmax()
    else:
        start_indx = x.start or 0

    if isinstance(x.stop, Var):
        stop_indx = (colnames == x.stop.name).idxmax() + 1
    elif isinstance(x.stop, str):
        stop_indx = (colnames == x.stop).idxmax() + 1
    else:
        stop_indx = x.stop or len(colnames)

    if start_indx > stop_indx:
        return stop_indx, start_indx
    else:
        return start_indx, stop_indx

def var_put_cols(name, var, cols):
    if isinstance(name, list) and var.alias is not None:
        raise Exception("Cannot assign name to multiple columns")
    
    names = [name] if not isinstance(name, list) else name

    for name in names:
        if var.negated:
            if name in cols: cols.pop(name)
        #elif name in cols: cols.move_to_end(name)
        else: cols[name] = var.alias

def flatten_var(var):
    if isinstance(var, VarAnd):
        return var.flatten()
    return [var]


def var_select(colnames, *args, data=None):
    # TODO: don't erase named column if included again
    colnames = colnames if isinstance(colnames, pd.Series) else pd.Series(colnames)
    cols = OrderedDict()

    #flat_args = var_flatten(args)
    all_vars = chain(*map(flatten_var, args))

    # Add entries in pandas.rename style {"orig_name": "new_name"}
    for ii, arg in enumerate(all_vars):

        # strings are added directly
        if isinstance(arg, str):
            cols[arg] = None

        # integers add colname at corresponding index
        elif isinstance(arg, int):
            cols[colnames.iloc[arg]] = None

        # general var handling
        elif isinstance(arg, Var):
            # remove negated Vars, otherwise include them
            if ii == 0 and arg.negated:
                # if negation used as first arg apply an implicit everything
                cols.update((k, None) for k in colnames)

            # slicing can refer to single, or range of columns
            if isinstance(arg.name, slice):
                start, stop = var_slice(colnames, arg.name)
                for ii in range(start, stop):
                    var_put_cols(colnames[ii], arg, cols)

            # method calls like endswith()
            elif callable(arg.name):
                # TODO: not sure if this is a good idea...
                #       basically proxies to pandas str methods (they must return bool array)
                indx = arg.name(colnames.str)
                var_put_cols(colnames[indx].tolist(), arg, cols)
                #cols.update((x, None) for x in set(colnames[indx]) - set(cols))
            elif isinstance(arg.name, int):
                var_put_cols(colnames.iloc[arg.name], arg, cols)
            else:
                var_put_cols(arg.name, arg, cols)
        elif callable(arg) and data is not None:
            # TODO: call on the data
            col_mask = colwise_eval(data, arg)

            for name in colnames[col_mask]:
                cols[name] = None


        else:
            raise Exception("variable must be either a string or Var instance")

    return cols


def var_create(*args) -> "tuple[Var]":
    vl = VarList()
    all_vars = []
    for arg in args:
        if isinstance(arg, Call):
            res = arg(vl)
            if isinstance(res, VarList):
                raise ValueError("Must select specific column. Did you pass `_` to select?")
            all_vars.append(res)
        elif isinstance(arg, Var):
            all_vars.append(arg)
        elif callable(arg):
            all_vars.append(arg)
        else:
            all_vars.append(Var(arg))
     
    return tuple(all_vars)


@singledispatch
def colwise_eval(data, predicate):
    raise NotImplementedError(
        f"Cannot evaluate tidyselect predicate on data type: {type(data)}"
    )


@colwise_eval.register
def _colwise_eval_pd(data: pd.DataFrame, predicate) -> List[bool]:
    mask = []
    for col_name in data:
        res = predicate(data.loc[:, col_name])
        if not pd.api.types.is_bool(res):
            raise TypeError("TODO")

        mask.append(res)

    return mask

    
    



"""GroupByAgg class and generic methods for fast pandas grouped operations.
"""

import inspect

from functools import singledispatch

from pandas import Series
from pandas.api.types import is_scalar
from pandas.core.groupby import SeriesGroupBy, DataFrameGroupBy

try:
    from pandas.core.algorithms import take_1d
except ImportError:
    from pandas.core.array_algos.take import take_1d


# Custom SeriesGroupBy class ==================================================

class GroupByAgg(SeriesGroupBy):
    """Class for representing the result of a grouped Series aggregation.

    Imagine that you are trying to add two grouped Series, where one might be
    an aggregate:

    >>> from siuba.data import mtcars
    >>> g_cyl = mtcars.groupby("cyl")
             
    >>> avg_hp_raw = g_cyl.hp.mean()
    >>> # how can we do: g_cyl.hp - avg_hp_raw ?

    This class is designed to allows operations like minute (``-``) to work under 3 cases:

    * original - aggregate: broadcast to original length; return SeriesGroupBy.
    * aggregate - aggregate: no need to broadcast; return GroupByAgg.
    * unary method over aggregate: no need to broadcast; return GroupByAgg.

    Due to complexities in how pandas creates grouped objects, the easiest way
    to create this class is to use its ``from_result`` class method:

    >>> avg_hp = GroupByAgg.from_result(avg_hp_raw, g_cyl.hp)
    
    Below are examples of the first two cases:

    >>> # avg_hp plus hp ----
    >>> x, y, grp = broadcast_group_elements(avg_hp, g_cyl.hp)
    >>> res1 = regroup(grp, x + y)           # SeriesGroupBy
    
    >>> # avg_hp plus avg_hp ----
    >>> x, y, grp = broadcast_group_elements(avg_hp, avg_hp)
    >>> res2 = regroup(grp, x + y)    # GroupByAgg

    You can use `is_compatible` to check whether broadcasting will work:
    
    >>> is_compatible(g_cyl.hp, avg_hp)
    True
    
    Lastly, this is a subclass of SeriesGroupBy, where each row is its
    own group, so unary methods can be performed without issue.:

    >>> res3 = regroup(grp, avg_hp.fillna(1))  # GroupByAgg
    >>> ser = broadcast_agg(res3)           # ungroup, same len as original data

    """


    def __init__(self, *args, orig_grouper, orig_obj, **kwargs):
        self._orig_grouper = orig_grouper
        self._orig_obj = orig_obj
        super().__init__(*args, **kwargs)
    
    @classmethod
    def from_result(cls, result: Series, src_groupby: SeriesGroupBy):
        """GroupByAgg class constructor.
        """

        if not isinstance(result, Series):
            raise TypeError("requires pandas Series")

        # Series.groupby is hard-coded to produce a SeriesGroupBy,
        # but its signature is very large, so use inspect to bind on it.
        sig = inspect.signature(result.groupby)
        bound = sig.bind(by = result.index)
        
        orig_grouper = getattr(src_groupby, "_orig_grouper", src_groupby.grouper)
        orig_obj     = getattr(src_groupby, "_orig_obj", src_groupby.obj)
        
        return cls(
            result,
            *bound.args, **bound.kwargs,
            orig_grouper = orig_grouper,
            orig_obj = orig_obj,
            )


# SeriesGroupBy generic functions =============================================

# broadcast_agg ---- 

@singledispatch
def broadcast_agg(groupby, result, obj):
    """Return a Series that is the length of original data, with original index."""

    raise NotImplementedError()


@broadcast_agg.register(GroupByAgg)
def _broadcast_agg_gba(groupby):
    """
    >>> import pandas as pd
    >>> gdf = pd.DataFrame({"g": ['a','a','b'], "x": [4,5,6]}).groupby("g")
    >>> agg = GroupByAgg.from_result(gdf.x.mean(), gdf.x)

    >>> len(broadcast_agg(agg))
    3

    """

    src = groupby._orig_obj
    ids, _, ngroup = groupby._orig_grouper.group_info
    out = take_1d(groupby.obj._values, ids)
    
    # Note: reductions like siuba.dply.vector.n(_) map DataFrameGroupBy -> GroupByAgg,
    # so the underlying object is a DataFrame, and does not have a .name attribute.
    return Series(out, index=src.index, name=getattr(src, "name", None))

@broadcast_agg.register(SeriesGroupBy)
def _broadcast_agg_sgb(groupby):
    return groupby.obj


# regroup ----

@singledispatch
def regroup(groupby, res):
    """Return an instance of type(groupby) from res."""

    raise TypeError("Not implemented for group by class: %s"% type(groupby))

@regroup.register(GroupByAgg)
def _(groupby, res):
    return groupby.from_result(res, groupby)

@regroup.register(SeriesGroupBy)
def _(groupby, res):
    # TODO: this will always return SeriesGroupBy, even if groupby is a subclass
    return res.groupby(groupby.grouper)


# is_compatible ----

@singledispatch
def is_compatible(grp1: SeriesGroupBy, grp2: SeriesGroupBy):
    """Return whether objects have identical original groupers."""

    grouper1 = getattr(grp1, '_orig_grouper', grp1.grouper)
    grouper2 = getattr(grp2, '_orig_grouper', grp2.grouper)

    return grouper1 is grouper2


# Utils =======================================================================

def all_isinstance(cls, *args):
    return all(isinstance(x, cls) for x in args)

# Broadcasting Groupby elements -----------------------------------------------

def grouper_match(grp1: SeriesGroupBy, grp2):
    # No need to broadcast against a scalar (pandas will handle) ----
    if is_scalar(grp2):
        return grp1.obj, grp2
    
    # Broadcasting requires: non-agg groupby with same original grouper ----
    if not isinstance(grp2, SeriesGroupBy):
        raise TypeError("grp2 must be a scalar or SeriesGroupBy")

    if not is_compatible(grp1, grp2):
        raise ValueError("groups must have matching groupers")

    return broadcast_agg(grp1), broadcast_agg(grp2)
    

def broadcast_group_elements(x, y):
    """Returns 3-tuple of same-length x, y Series, plus a reference group by object.

    Note:
        * Raises error if x and y are not compatible group by objects.
        * Will broadcast a GroupByAgg, to ensure same length as other data.
    """

    # Both are aggregations, so don't need to broadcast ----
    if all_isinstance(GroupByAgg, x, y) and is_compatible(x, y):
        return x.obj, y.obj, x
    
    # Only one is an aggregation, may broadcast along other ----
    elif isinstance(x, SeriesGroupBy):
        res_x, res_y = grouper_match(x, y)
        return res_x, res_y, x
    
    elif isinstance(y, SeriesGroupBy):
        # same as above, but with args / results flipped
        res_y, res_x, grouper_match(y, x)
        return res_x, res_y, y
    
    # Both are non-agg groupby, just need underlying objects ----
    #elif all_isinstance(SeriesGroupBy, x, y) and is_compatible(x, y):
    #    return x.obj, y.obj, x
    

    raise ValueError("need scalar, or groupby objects with matching groupers")




import pytest
from siuba.tests.helpers import data_frame
import pandas as pd

from siuba.experimental.pd_groups.translate import method_agg_op, method_el_op, method_el_op2
from siuba.experimental.pd_groups.groupby import broadcast_agg
#TODO: 
#  - what if they have mandatory, non-data args?
#  - support accessor methods like _.x.str.upper()
#  - support .expanding and .rolling


data_dt = data_frame(
    g = ['a', 'a', 'b', 'b'],
    x = pd.to_datetime(["2019-01-01 01:01:01", "2020-04-08 02:02:02", "2021-07-15 03:03:03", "2022-10-22 04:04:04"])
    )

data_str = data_frame(
    g = ['a', 'a', 'b', 'b'],
    x = ['abc', 'cde', 'fg', 'h']
    )

data_default = data_frame(
    g = ['a', 'a', 'b', 'b'],
    x = [10, 11, 12, 13],
    y = [1,2,3,4]
    )

data = {
    'dt': data_dt,
    'str': data_str,
    None: data_default
}


# Test translator =============================================================

from pandas.testing import assert_frame_equal, assert_series_equal
from siuba.experimental.pd_groups.groupby import GroupByAgg, SeriesGroupBy

f_min = method_agg_op('min', is_property = False, accessor = None)
f_add = method_el_op2('add', is_property = False, accessor = None)
f_abs = method_el_op('abs', is_property = False, accessor = None)
f_df_size = lambda x: GroupByAgg.from_result(x.size(), x)

# GroupByAgg is liskov substitutable, so check that our functions operate
# like similarly substitutable subtypes. This means that...
# * input type is the same or more general, and
# * output type is the same or more specific

@pytest.mark.parametrize('f_op, f_dst, cls_result', [
        # aggregation 1-arity
        # f(SeriesGroupBy) -> GroupByAgg <= f(GroupByAgg) -> GroupByAgg 
        (lambda g: f_min(g.x),        lambda g: g.x.min(), GroupByAgg),
        (lambda g: f_min(f_min(g.x)), lambda g: g.x.min(), GroupByAgg),
        # elementwise 1-arity
        # f(GroupByAgg) -> GroupByAgg <= f(SeriesGroupBy) -> SeriesGroupBy 
        (lambda g: f_abs(f_min(g.x)), lambda g: g.x.min().abs(), GroupByAgg),
        (lambda g: f_abs(g.x),        lambda g: g.obj.x.abs(), SeriesGroupBy),
        # elementwise 2-arity
        # f(GroupByAgg, GroupByAgg) -> GroupByAgg <= f(GroupByAgg, SeriesGroupBy) -> SeriesGroupBy
        (lambda g: f_add(f_min(g.x), f_min(g.y)), lambda g: g.x.min() + g.y.min(), GroupByAgg),
        (lambda g: f_add(g.x, f_min(g.y)),        lambda g: g.obj.x + g.y.transform('min'), SeriesGroupBy),
        (lambda g: f_add(g.x, g.y),               lambda g: g.obj.x + g.obj.y, SeriesGroupBy),
        ])
def test_grouped_translator_methods(f_op, f_dst, cls_result):
    g = data_default.groupby('g')
    res = f_op(g)

    # needs to be exact, since GroupByAgg is subclass of SeriesGroupBy
    assert type(res) is cls_result

    dst = f_dst(g)
    assert_series_equal(res.obj, dst, check_names = False)


@pytest.mark.parametrize('f_op, f_dst', [
        (lambda g: f_add(f_min(g.x), f_min(g.y)), lambda g: g.x.transform('min') + g.y.transform('min')),
        (lambda g: f_min(g.x),        lambda g: g.x.transform('min')),
        (lambda g: f_min(f_min(g.x)), lambda g: g.x.transform('min')),
        (lambda g: f_abs(f_min(g.x)), lambda g: g.x.transform('min').abs()),

        # Note that there's no way to transform a DF method, so use an arbitrary column
        (lambda g: f_df_size(g), lambda g: g.x.transform('size')),
        ])
def test_agg_groupby_broadcasted_equal_to_transform(f_op, f_dst):
    g = data_default.groupby('g')
    res = f_op(g)

    # needs to be exact, since GroupByAgg is subclass of SeriesGroupBy
    assert type(res) is GroupByAgg

    dst = f_dst(g)
    broadcasted = broadcast_agg(res)
    assert_series_equal(broadcasted, dst, check_names = False)


# Test generic functions ======================================================

def test_fast_mutate_basic():
    # sanity check of https://github.com/machow/siuba/issues/355
    from siuba.siu import _

    res_df = data_default.groupby("g") >> fast_mutate(num = _.x / _.y * 100)

    res = res_df.num
    dst = data_default.x / data_default.y * 100

    assert_series_equal(res.obj, dst, check_names=False)
    

# Test user-defined functions =================================================

from .dialect import fast_mutate, fast_summarize, fast_filter, _transform_args
from siuba.siu import symbolic_dispatch, _, FunctionLookupError
from typing import Any

def test_transform_args():
    pass


def test_fast_grouped_custom_user_funcs():
    @symbolic_dispatch
    def f(x):
        raise NotImplementedError()

    @f.register(SeriesGroupBy)
    def _f_grouped(x) -> GroupByAgg:
        return GroupByAgg.from_result(x.mean() + 10, x)

    gdf = data_default.groupby('g')
    g_out = fast_mutate(gdf, result1 = f(_.x), result2 = _.x.mean() + 10)
    out = g_out.obj
    assert (out.result1 == out.result2).all()


def test_fast_grouped_custom_user_func_default():
    @symbolic_dispatch
    def f(x) -> Any:
        return GroupByAgg.from_result(x.mean() + 10, x)

    gdf = data_default.groupby('g')
    g_out = fast_mutate(gdf, result1 = f(_.x), result2 = _.x.mean() + 10)
    out = g_out.obj
    assert (out.result1 == out.result2).all()

def test_fast_grouped_custom_user_func_fail():
    @symbolic_dispatch
    def f(x):
        return x.mean()

    @f.register(SeriesGroupBy)
    def _f_gser(x):
        # note, no return annotation, so translator will raise an error
        return GroupByAgg.from_result(x.mean(), x)


    gdf = data_default.groupby('g')
    
    with pytest.warns(UserWarning):
        g_out = fast_mutate(gdf, result1 = f(_.x), result2 = _.x.mean() + 10)


def test_fast_methods_constant():
    gdf = data_default.groupby('g')

    # mutate ----
    out = fast_mutate(gdf, y = 1)
    assert_frame_equal(data_default.assign(y = 1), out.obj)

    # summarize ----
    out = fast_summarize(gdf, y = 1)
    agg_frame = gdf.grouper.result_index.to_frame().reset_index(drop = True)
    assert_frame_equal(agg_frame.assign(y = 1), out)

    # filter ----
    out = fast_filter(gdf, True)
    assert_frame_equal(gdf.obj, out.obj)

    # note: two empty DataFrames can differ if their (empty) indices are different types
    #       e.g. Index vs RangeIndex
    out = fast_filter(gdf, False)
    assert_frame_equal(
            gdf.obj.iloc[0:0,:],
            out.obj,
            check_index_type = False
            )

    
def test_fast_methods_lambda():
    # testing ways to do operations via slower apply route

    gdf = data_default.groupby('g')

    # mutate ----
    out = fast_mutate(gdf, y = lambda d: len(d['x']))
    assert_frame_equal(
            gdf.obj.assign(y = gdf['x'].transform('size')),
            out.obj
            )

    # summarize ----
    out = fast_summarize(gdf, y = lambda d: len(d['x']))

    agg_frame = gdf.grouper.result_index.to_frame()
    assert_frame_equal(
            agg_frame.assign(y = gdf['x'].agg('size')).reset_index(drop = True),
            out
            )

    # filter ----
    out = fast_filter(gdf, lambda d: d['x'] > d['x'].values.min())
    assert_frame_equal(
            gdf.obj[gdf.obj['x'] > gdf['x'].transform('min')],
            out.obj
            )



from sqlalchemy.sql import func as fn
from sqlalchemy import sql

from ..translate import (
    # data types
    SqlColumn, SqlColumnAgg,
    AggOver,
    # transformations
    wrap_annotate,
    sql_agg,
    win_agg,
    win_cumul,
    sql_not_impl,
    # wiring up translator
    extend_base,
    SqlTranslator,
    convert_literal
)

from .postgresql import (
    PostgresqlColumn,
    PostgresqlColumnAgg,
)

from .base import sql_func_rank


# Data ========================================================================

class DuckdbColumn(PostgresqlColumn): pass
class DuckdbColumnAgg(PostgresqlColumnAgg, DuckdbColumn): pass


# Annotations =================================================================

def returns_int(func_names):
    # TODO: MC-NOTE - shift all translations to directly register
    # TODO: MC-NOTE - make an AliasAnnotated class or something, that signals
    #                 it is using another method, but w/ an updated annotation.
    from siuba.ops import ALL_OPS
    
    for name in func_names:
        generic = ALL_OPS[name]
        f_concrete = generic.dispatch(SqlColumn)
        f_annotated = wrap_annotate(f_concrete, result_type="int")
        generic.register(DuckdbColumn, f_annotated)
    
# Literal Conversions =========================================================

@convert_literal.register
def _cl_duckdb(codata: DuckdbColumn, lit):
    from sqlalchemy.dialects.postgresql import array
    if isinstance(lit, list):
        return array(lit)
    
    return sql.literal(lit)

# Translations ================================================================


def sql_quantile(is_analytic=False):
    # Ordered and theoretical set aggregates
    sa_func = getattr(sql.func, "percentile_cont")

    def f_quantile(codata, col, q, *args):
        if args:
            raise NotImplementedError("Quantile only supports the q argument.")
        if not isinstance(q, (int, float)):
            raise TypeError("q argument must be int or float, but received: %s" %type(q))

        # as far as I can tell, there's no easy way to tell sqlalchemy to render
        # the exact text a dialect would render for a literal (except maybe using
        # literal_column), so use the classic sql.text.
        q_text = sql.text(str(q))

        if is_analytic:
            return AggOver(sa_func(sql.text(q_text)).within_group(col))

        return sa_func(q_text).within_group(col)

    return f_quantile


# scalar ----

extend_base(
    DuckdbColumn,
    **{
        "str.contains": lambda _, col, re: fn.regexp_matches(col, re),
        "str.title": sql_not_impl(),
    }
)

returns_int([
    "dt.day", "dt.dayofyear", "dt.days_in_month",
    "dt.daysinmonth", "dt.hour", "dt.minute", "dt.month",
    "dt.quarter", "dt.second", "dt.week",
    "dt.weekofyear", "dt.year"
])

# window ----

extend_base(
    DuckdbColumn,
    __floordiv__ = lambda _, x, y: x.op("//")(y),
    __rfloordiv__ = lambda _, y, x: x.op("//")(y),
    rank = sql_func_rank,
    #quantile = sql_quantile(is_analytic=True),
)


# aggregate ----

extend_base(
    DuckdbColumnAgg,
    quantile = sql_quantile(),
)


translator = SqlTranslator.from_mappings(DuckdbColumn, DuckdbColumnAgg)



from sqlalchemy import sql

from siuba.dply.verbs import summarize

from .mutate import _sql_upsert_columns, _eval_expr_arg, _eval_expr_kwarg

from ..utils import lift_inner_cols, _sql_with_only_columns
from ..backend import LazyTbl, get_single_from


@summarize.register(LazyTbl)
def _summarize(__data, *args, **kwargs):
    # https://stackoverflow.com/questions/14754994/why-is-sqlalchemy-count-much-slower-than-the-raw-query

    # get query with correct from clause, and maybe unneeded subquery
    safe_from = __data.last_select.alias()
    result_names, sel = _aggregate_cols(__data, safe_from, args, kwargs, "Summarize")

    # see if we can remove subquery
    out_sel = _collapse_select(sel, safe_from)

    from_tbl = get_single_from(out_sel)
    group_cols = [from_tbl.columns[k] for k in __data.group_by]

    final_sel = out_sel.group_by(*group_cols)

    new_data = __data.append_op(final_sel, group_by = tuple(), order_by = tuple())
    return new_data


def _collapse_select(outer_sel, inner_alias):
    # check whether any outer columns reference an inner label ----
    inner_sel = inner_alias.element

    columns = lift_inner_cols(outer_sel)
    inner_cols = lift_inner_cols(inner_sel)

    inner_labels = set([
        x.name for x in inner_cols 
        if isinstance(x, sql.elements.Label)
    ])

    col_requires_cte = set(inner_alias.columns[k] for k in inner_labels)

    bad_refs = []

    def collect_refs(el):
        if el in col_requires_cte:
            bad_refs.append(el)

    for col in columns:
        sql.util.visitors.traverse(col, {}, {"column": collect_refs})

    # if possible, remove the outer query ----
    if not (bad_refs or len(inner_sel._group_by_clause)):
        from sqlalchemy.sql.elements import ColumnClause, Label

        from_obj = get_single_from(inner_sel)
        adaptor = sql.util.ClauseAdapter(
            from_obj,
            adapt_on_names=True,
            include_fn=lambda c: isinstance(c, (ColumnClause, Label))
        )

        new_cols = []
        for col in columns:
            if isinstance(col, Label):
                res = adaptor.traverse(col.element).label(col.name)
                new_cols.append(res)

            else:
                new_cols.append(adaptor.traverse(col))
        #new_cols = list(map(adaptor.traverse, columns))

        return _sql_with_only_columns(inner_sel, new_cols)

    return outer_sel


def _aggregate_cols(__data, subquery, args, kwargs, verb_name):
    # cases:
    #   * grouping cols can not be overwritten (in dbplyr they can't be ref'd)
    #   * no existing labels referred to - can use same select
    #   * existing labels referred to - need 1 subquery tops
    #   * groups + summarize columns can replace everything

    def get_label_clauses(clause):
        out = []
        sql.util.visitors.traverse(clause, {}, {"label": lambda c: out.append(c)})

        return out

    def quote_varname(x):
        return f"`{x}`"

    def validate_references(arg_name, expr, verb_name):
        bad_varnames = get_label_clauses(expr)
        repr_names = ", ".join(map(quote_varname, [el.name for el in bad_varnames]))

        if not bad_varnames:
            return

        raise NotImplementedError(
            f"In SQL, you cannot refer to a column created in the same {verb_name}. "
            f"`{arg_name}` refers to columns created earlier: {repr_names}."
        )

    sel = subquery.select()

    final_cols = {k: subquery.columns[k] for k in __data.group_by}

    # handle args ----
    for ii, func in enumerate(args):
        cols_result = _eval_expr_arg(__data, sel, func, verb_name, window=False)
        
        for col in cols_result:
            validate_references(col.name, col.element, verb_name)
            final_cols[col.name] = col

        sel = _sql_upsert_columns(sel, cols_result)


    # handle kwargs ----
    for new_name, func in kwargs.items():
        labeled = _eval_expr_kwarg(__data, sel, func, new_name, verb_name, window=False)

        validate_references(labeled.name, labeled.element, verb_name)
        final_cols[new_name] = labeled

        sel = _sql_upsert_columns(sel, [labeled])

    return list(final_cols), _sql_with_only_columns(sel, list(final_cols.values()))



"""
DEPRECATED.

Note that this module was experimental, and created very early in siuba's development.
You should not rely on it for anything important.
"""

from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
from importlib.util import find_spec
import importlib
from types import ModuleType
import sys
import warnings

from functools import wraps
import pdb
import readline

warnings.warn(
    "The siuba.meta_hook module is DEPRECATED and will be removed in a future release."
)




class CallFinder(MetaPathFinder):
    def __init__(self, f):
        self.enabled = True
        self.f = f

    def find_spec(self, fullname, path, target = None):
        if not self.enabled:
            return None

        pkg, *subpkgs = fullname.split('.')
        if fullname.startswith("siuba.meta_hook") and len(subpkgs) > 1:
        #if pkg == 'meta_hook' and len(subpkgs):
            self.enabled = False
            spec = find_spec(".".join(subpkgs[1:]))
            self.enabled = True
            #spec.loader = CallLoader(self.f, spec.loader, spec)
            return ModuleSpec(fullname, CallLoader(self.f, spec))
        elif pkg == "meta_hook":
            pass

    #def invalidate_caches(self):
    #    pass

class CallLoader(Loader):
    def __init__(self, f, spec):
        self.f = f
        self.orig_spec = spec

    def create_module(self, spec):
        #self.orig_module = self.orig_spec.loader.create_module(self.orig_spec)
        self.orig_module = importlib.import_module(self.orig_spec.name)        
        #self.orig_module = self.orig_loader.create_module(spec)
        #if self.orig_module is None:
        #    self.orig_module = ModuleType(spec.name)

        #return self.orig_module
        return None


    def exec_module(self, module):
        #self.orig_loader.exec_module(self.orig_module)

        #for k,v in self.orig_module.__dict__.items():
        all_items = list(self.orig_module.__dict__.items())
        for k,v in all_items:
            if k.startswith('_'):
                module.__dict__[k] = v
            else:
                module.__dict__[k] = self.f(v)


from .siu import Symbolic, Call

def lazy_func(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return Symbolic(source = f)(*args, **kwargs)

    return wrapper


if sys.meta_path[0].__class__.__name__ == "CallFinder":
    sys.meta_path[0] = CallFinder(lazy_func)
else:
    sys.meta_path = [CallFinder(lazy_func)] + sys.meta_path

__path__ = ''




import pandas as pd
import numpy as np

from pandas.core.groupby import DataFrameGroupBy
from siuba.dply.tidyselect import var_create, var_select
from siuba.dply.verbs import gather, separate, extract, expand, complete
from siuba.siu import singledispatch2, Call
from siuba.dply.forcats import fct_inorder

from .utils import vec_as_names, reconstruct_tibble

from typing import Any


def _select_expr_slice(x: "tuple[str]") -> Call:
    from operator import getitem
    from siuba.siu import strip_symbolic, Symbolic

    return strip_symbolic(
        getitem(Symbolic(), x)
    )

def _tidy_select(__data, cols, arg_name):
    if cols is None:
        return {}

    var_list = var_create(cols)
    od = var_select(__data.columns, *var_list)

    # TODO: where in tidyselect package does this check happen?
    missing = set(od) - set(__data.columns)

    if missing:
        raise ValueError(
            f"{arg_name} must select columns present in the data. "
            f"Could not find these columns: {missing}"
        )

    return od

def _maybe_list(seq):
    if seq is None:
        return None

    return list(seq)


def _collapse_index_names(index, sep : "str | None" = None, glue = None):
    if glue is not None:
        if index.nlevels == 1:
            tmp_index = [[x, ""] for x in index]
        else:
            tmp_index = index

        return [glue.format(variable=entry[0], value=entry[1]) for entry in tmp_index]
    elif sep is not None:
        if index.nlevels == 1:
            return list(index)
        return [sep.join(map(str, entry)) for entry in index]
    else:
        raise NotImplementedError()


def _unique_col_name(df):
    base = "__TMP_COL__"
    test = base
    ii = 0

    df_cols = set(df.columns)
    while test in df.columns:
        test = base + str(ii)
        ii += 1

    return test


def _is_select_everything(expr):
    # this is crazy but probably fine for now.
    if (isinstance(expr, Call) and expr.func == "__getitem__"):
        sub_expr = expr.args[1]
        return (
            isinstance(sub_expr, Call)
            and sub_expr.func == "__siu_slice__"
            and len(sub_expr.args) == 1 and
            sub_expr.args[0] == slice(None, None, None)
        )
    return False


def _names_from_spec(spec, multi_index):
    # dict mapping name parts to the final names
    d_spec = dict(zip(spec.iloc[:, 1:].itertuples(index=False, name=None), spec[".name"]))

    return [d_spec[x] for x in multi_index]


@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def pivot_wider(
    __data,
    id_cols=None,
    id_expand=False,
    names_from="name",
    names_prefix="",
    names_sep="_",
    names_glue=None,
    names_sort=None,
    names_vary="fastest",
    names_expand=False,
    names_repair="check_unique",
    values_from="value",
    values_fill=None,
    values_fn=None,
    unused_fn=None
    ):
    """Pivot data from long to wide format.

    This function splits a column, putting the pieces side-by-side based on an index.

    Parameters
    ----------
    __data:
        The input data.
    id_cols:
        A selection of columns that uniquely identify each observation.
    id_expand:
        Whether to ensure each unique combination of id_cols is a row in the data
        before pivoting, using `expand()`. This results in more rows. When True,
        this also sorts the final result by the `id_cols`.
    names_from, values_from:
        A pair fo arguments describing which column (or columns) to get the name of
        the output column (names_from), and which column (or columns) to get the
        cell values from (values_from).
    names_prefix:
        String added to the start of every variable name.
    names_sep:
        If names_from or values_from contains multiple values, this will be used
        to join their values together into a single string to use as a column name.
    names_glue:
        Instead of names_sep and names_prefix, supply a string template that uses
        the names_from columns (and a special .value variable) to create custom
        column names.
    names_sort:
        Should the column names be sorted? The default is False, which results
        in column names ordered by first appearance.
    names_vary:
        Option specifying how columns are ordered when names_from and values_from
        both identify new columns. "fastest" varies names_from fastest, while "slowest"
        varies names_from slowest.
    names_expand:
        Whether to ensure all combinations of names_from columns are in the result
        using the `expand()` function. This results in more columns in the output.
    names_repair:
        Strategy for fixing of invalid column names. "minimal" leaves them as is.
        "check_unique" raises an error if there are duplicate names. "unique"
        de-duplicates names by appending "___{position}" to them.
    values_fill:
        A scalar value used to fill in any missing values. Alternatively, a
        dictionary mapping column names to fill values.
    values_fn:
        An optional function to apply to each cell of the output. This is useful
        when each cell would contain multiple values. E.g. values_fn="max" would
        calculate the max value.
    unused_fn:
        Not implemented.

    Examples
    --------
    >>> from siuba import _

    >>> df = pd.DataFrame(
    ...    {"id": ["a", "b", "a"], "name": ["x", "x", "y"], "value": [1, 2, 3]}
    ... )
    >>> df
      id name  value
    0  a    x      1
    1  b    x      2
    2  a    y      3

    >>> pivot_wider(df, names_from=_.name, values_from=_.value)
      id    x    y
    0  a  1.0  3.0
    1  b  2.0  NaN

    >>> pivot_wider(df, names_from=_.name, values_from=_.value, values_fill=0)
      id  x  y
    0  a  1  3
    1  b  2  0

    >>> many = pd.DataFrame({
    ...    "id": [1, 1, 2, 2],
    ...    "var": ["one", "two", "one", "two"],
    ...    "x": [1, 2, 3, 4],
    ...    "y": [6, 7, 8, 9]
    ... })
    >>> pivot_wider(many, names_from=_.var, values_from=_[_.x, _.y])
       id  x_one  x_two  y_one  y_two
    0   1      1      2      6      7
    1   2      3      4      8      9

    >>> pivot_wider(many, names_from=_.var, values_from=_[_.x, _.y], names_vary="slowest")
       id  x_one  y_one  x_two  y_two
    0   1      1      6      2      7
    1   2      3      8      4      9

    >>> pivot_wider(many, names_from=_.var, values_from=_[_.x, _.y], names_sep=".")
       id  x.one  x.two  y.one  y.two
    0   1      1      2      6      7
    1   2      3      4      8      9

    >>> glue = "{variable}_X_{value}"
    >>> pivot_wider(many, names_from=_.var, values_from=_[_.x, _.y], names_glue=glue)
       id  x_X_one  x_X_two  y_X_one  y_X_two
    0   1        1        2        6        7
    1   2        3        4        8        9

    >>> from siuba.data import warpbreaks
    >>> warpbreaks.head()
       breaks wool tension
    0      26    A       L
    1      30    A       L
    2      54    A       L
    3      25    A       L
    4      70    A       L

    >>> pivot_wider(warpbreaks, names_from=_.wool, values_from=_.breaks, values_fn="mean")
      tension          A          B
    0       H  24.555556  18.777778
    1       L  44.555556  28.222222
    2       M  24.000000  28.777778

    """
    input_ = __data

    if isinstance(__data, DataFrameGroupBy):
        __data = __data.obj

    # create spec ----
    spec = build_wider_spec(
        __data,
        names_from = names_from,
        values_from = values_from,
        names_prefix = names_prefix,
        names_sep = names_sep,
        names_glue = names_glue,
        names_sort = names_sort,
        names_vary = names_vary,
        names_expand = names_expand
    )

    # select id columns ---
    # necessary here, since if the spec is 0 rows you cannot know values_from
    # TODO: clean up symbolic handling of slices
    if id_cols is None:
        name_vars = _tidy_select(__data, names_from, "names_from")
        val_vars = _tidy_select(__data, values_from, "values_from")
        others = {*name_vars, *val_vars}

        id_cols = tuple([col for col in __data.columns if col not in others])
        id_vars = _select_expr_slice(id_cols)
    else:
        id_vars = id_cols

    out = pivot_wider_spec(
        input_,
        spec,
        names_repair = names_repair,
        id_cols = id_vars,
        id_expand = id_expand,
        values_fill = values_fill,
        values_fn = values_fn,
        unused_fn = unused_fn
    )

    return out

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def pivot_wider_spec(
    __data,
    spec,
    names_repair = "check_unique",
    id_cols = None,
    id_expand = False,
    values_fill = None,
    values_fn = None,
    unused_fn = None
):

    input_ = __data

    # guards ----

    if isinstance(__data, DataFrameGroupBy):
        __data = __data.obj

    if _is_select_everything(id_cols):
        # restores id_cols to the default, which uses all remaining cols
        id_cols = None

    if unused_fn is not None:
        raise NotImplementedError()

    if not isinstance(id_expand, bool):
        raise TypeError("`id_expand` argument must be True or False.")

    # handle tidyselection ----------------------------------------------------

    name_vars = spec.columns[~spec.columns.isin([".name", ".value"])].tolist()
    val_vars = spec.loc[:, ".value"].unique().tolist()

    # select id columns
    if id_cols is None:
        others = {*name_vars, *val_vars}
        id_vars = [col for col in __data.columns if col not in others]
    else:
        id_vars = _tidy_select(__data, id_cols, "id_cols")

    id_var_bad = set(id_vars) & set([*name_vars, *val_vars])
    if id_var_bad:
        raise ValueError(
            "id_cols contains columns that are in "
            f"names_from or values_from: {id_var_bad}."
        )

    # use a categoricals for name columns, to ensure their order in wide format
    # is first-observed order (pandas by default uses alphabetical)
    tmp = __data.copy()
    for name in name_vars:
        tmp[name] = fct_inorder(tmp[name])

    # pivot to wide -----------------------------------------------------------

    if values_fn is None:
        # this block is essentially pd.pivot (which also uses unstack), but also
        # supports filing NAs, and resets indexes
        if not len(id_vars):
            # without id_vars we try to pivot to a frame with 1 row
            n_id_vars = 1
            tmp = tmp.set_index([np.zeros(len(tmp.index)), *name_vars])
        else:
            n_id_vars = len(id_vars)
            tmp = tmp.set_index([*id_vars, *name_vars])

        to_unstack = list(range(n_id_vars, n_id_vars + len(name_vars)))
        wide = (tmp
            .loc[:, list(val_vars)]
            .unstack(to_unstack, fill_value=values_fill)
        )

    else:
        # pivot_table requires a values_fn, so we only use it when one is provided.
        if not len(id_vars):
            # pivot_table without an index var is equivalent to the index being constant.
            # normally the index vars are the index of the pivot_table result, but without
            # an index column explicitly named, the value vars become the rows.
            # so we need to create an explicit index column...
            index_cols = [_unique_col_name(tmp)]
            tmp.loc[:, index_cols[0]] = np.zeros(len(tmp.index))

        else:
            index_cols = list(id_vars)

        # this ensures a single value column won't be used when constructing names
        # since a list creates a new index dimension
        _values = list(val_vars) if len(val_vars) > 1 else list(val_vars)[0]

        wide = pd.pivot_table(
            tmp,
            index=index_cols,
            columns=list(name_vars),
            values=list(val_vars),
            fill_value=values_fill,
            aggfunc=values_fn,
        )

        if wide.index.names != index_cols:
            raise ValueError(
                "pivot_wider produced a result with incorrect index variables. "
                "There is a bug in pandas when attempting to aggregate by a values_fn "
                "that is not an aggregate.\n\n"
                f"Do all the values_fn arguments return single values?: {values_fn}"
            )


    # flatten / reset indexes -------------------------------------------------

    # flatten column index ----
    if isinstance(wide, pd.Series):
        # the .unstack approach returns a Series when there are no id cols.
        # in this case we make it a frame and don't sort the two columns.
        wide = wide.reset_index()

    collapsed_names = _names_from_spec(spec, wide.columns)
    wide.columns = collapsed_names

    # add missing columns and reorder to spec ----
    missing_cols = list(spec[".name"][~spec[".name"].isin(wide.columns)])

    if missing_cols:
        wide[missing_cols] = values_fill

    wide = wide.loc[:, list(spec[".name"])]

    # validate names and move id vars to columns ----
    # note: in pandas 1.5+ we can use the allow_duplicates option to reset, even
    # when index and column names overlap. for now, repair names, rename, then reset.
    _all_raw_names = list(map(str, [*id_vars, *wide.columns]))
    unique_names = vec_as_names(_all_raw_names, repair="unique")
    repaired_names = vec_as_names(_all_raw_names, repair=names_repair)

    uniq_id_vars = unique_names[:len(id_vars)]
    uniq_val_vars = unique_names[len(id_vars):]

    final_id_vars = repaired_names[:len(id_vars)]
    final_val_vars = repaired_names[len(id_vars):]

    wide.columns = uniq_val_vars

    if id_vars:
        wide.rename_axis(uniq_id_vars, inplace=True)
        wide.reset_index(drop=False, inplace=True)
    else:
        wide.reset_index(drop=True, inplace=True)

    wide.columns = repaired_names

    # expand id levels --------------------------------------------------------

    if id_expand:
        if values_fill is not None and not isinstance(values_fill, dict):
            values_fill = {k: values_fill for k in final_val_vars}

        wide = complete(wide, *final_id_vars, fill=values_fill, explicit=False)
        
    # reconstruct with groupings
    return reconstruct_tibble(input_, wide)


@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def build_wider_spec(
    __data,
    names_from = "name",
    values_from = "value",
    names_prefix = "_",
    names_sep = "_",
    names_glue = None,
    names_sort = False,
    names_vary = "fastest",
    names_expand = False
):
    if isinstance(__data, DataFrameGroupBy):
        __data = __data.obj

    # guards ----
    if names_vary not in {"fastest", "slowest"}:
        raise ValueError(
            "names_vary must be one of 'fastest', 'slowest', but received "
            f"argument: {repr(names_vary)}"
        )

    if names_sort:
        raise NotImplementedError()

    if not isinstance(names_expand, bool):
        raise TypeError(
            "names_expand must be set to True or False. "
            f"Received type: {type(names_expand)}."
        )

    # validate tidy selections ------------------------------------------------
    orig_vars = __data.columns

    name_vars = _tidy_select(__data, names_from, "names_from")
    val_vars = _tidy_select(__data, values_from, "values_form")

    if not name_vars:
        raise ValueError("`names_from` must select at least one column.")

    if not val_vars:
        raise ValueError("`values_from` must select at least one column.")

    # get unique variable levels from names_from columns ----------------------

    name_data = __data.loc[:, list(name_vars)]
    if names_expand:
        # cartesian product of unique level names
        # TODO: should nan values be turned into "NA" to match dplyr?
        row_ids = expand(name_data, *name_vars)
    else:
        # distinct rows of variables
        row_ids = name_data.drop_duplicates()

    # cross with value var names ----------------------------------------------

    value_levels = pd.Series(list(val_vars), name = ".value")
    if names_vary == "fastest":
        spec = pd.merge(
            value_levels,
            row_ids,
            how="cross"
        )
    else:
        # the left arg varies slowest, so use names on the left, then relocate.
        spec = (
            pd.merge(
                row_ids,
                value_levels,
                how="cross"
            )
            .loc[:, lambda d: [".value", *d.columns[:-1]]]
        )

    # get columns used to construct .name
    if len(value_levels) > 1:
        df_name_parts = spec
    else:
        df_name_parts = spec.drop(columns=".value")

    # TODO: remove use of multiindex, which is unnecessary
    if len(df_name_parts.columns) > 1:
        spec_as_multi = pd.MultiIndex.from_frame(df_name_parts)
        name_col = _collapse_index_names(spec_as_multi, sep=names_sep, glue=names_glue)
    else:
        name_col = list(df_name_parts.iloc[:, 0])

    spec.insert(0, ".name", name_col)

    return spec



import pandas as pd
import numpy as np
import re

from typing import Union, Tuple, Dict, Optional, Callable
from pandas.core.groupby import DataFrameGroupBy

from siuba.dply.verbs import gather, var_create, var_select, separate, extract
from siuba.siu import singledispatch2

from .utils import vec_as_names, reconstruct_tibble, check_dict_of_functions

# Utilities ===================================================================

def spec_to_multiindex(df_spec):
    # _value will be the outer column index, and split name columns the inner.
    # this allows us to stack the inner part, while that _values stays as columns.
    internal = {".value", ".name"}

    #other_cols = [name for name in df_spec.columns if name not in internal]
    #df_spec[other_cols].apply(lambda ser: tuple(ser), axis=1)
    other_cols = [nm for nm in df_spec.columns if nm not in internal]

    # get final columns together
    indx_cols = [".value", *other_cols]
    indx_names = [None] + other_cols

    df_final = df_spec.loc[:, indx_cols]

    # ensure levels of _value in multi-index are in first-observed order
    # otherwise, their final columns come out alphabetically in pivot_longer.
    val = df_final[".value"]
    df_final[".value"] = pd.Categorical(val, val.dropna().unique())

    return pd.MultiIndex.from_frame(df_final, names=indx_names)


def _drop_cols_by_position(df, locations):
    return df.loc[:, ~np.array([x in locations for x in range(df.shape[1])])]


# Pivot longer ================================================================

@singledispatch2(pd.DataFrame)
def pivot_longer(
        __data,
        *cols,
        names_to: Union[str, Tuple[str, ...]] = "name",
        names_prefix: Optional[str] = None,
        names_sep: Optional[str] = None,
        names_pattern: Optional[str] = None,
        names_ptypes: Optional[Tuple] = None,
        names_repair: str = "check_unique",
        values_to: str = "value",
        values_drop_na: bool = False,
        values_ptypes: Optional[Union[str, Tuple[str, ...]]] = None,
        values_transform: Optional[Dict] = None,
):
    """Pivot data from wide to long format.

    This function stacks columns of data, turning them into rows.

    Parameters
    ----------
    __data:
        The input data.
    *cols: 
        Columns to pivot into longer format. This uses tidyselect
        (e.g. `_[_.some_col, _.another_col]`).
    names_to:
        A list specifying the new column or columns to create from the information
        stored in the column names of data specified by cols.
    names_prefix:
        A regular expression to strip off from the start of column selected by `*cols`.
    names_sep:
        If names_to is a list of name parts, this is a separater the name is split on.
        This is the same as the sep argument in the separate() function.
    names_pattern:
        If names_to is a list of name parts, this is a pattern to extract parts
        This is the same as the regex argument in the extract() function.
    names_ptypes, values_ptypes:
        Not implemented.
    names_transform:
        TODO
    names_repair:
        Strategy for fixing of invalid column names. "minimal" leaves them as is.
        "check_unique" raises an error if there are duplicate names. "unique"
        de-duplicates names by appending "___{position}" to them.
    values_to:
        A string specifying the name of the column created to hold the stacked
        values of the selected `*cols`. If names_to is a list with the entry ".value",
        then this argument is ignored.


    Examples
    --------
    >>> from siuba import _

    >>> df = pd.DataFrame({"id": [1, 2], "x": [5, 6], "y": [7, 8]})
    >>> pivot_longer(df, ~_.id, names_to="variable", values_to="number")
       id variable  number
    0   1        x       5
    0   1        y       7
    1   2        x       6
    1   2        y       8

    >>> weeks = pd.DataFrame({"id": [1], "year": [2020], "wk1": [5], "wk2": [6]})
    >>> pivot_longer(weeks, _.startswith("wk"), names_to="week", names_prefix="wk")
       id  year week  value
    0   1  2020    1      5
    0   1  2020    2      6

    >>> df2 = pd.DataFrame({"id": [1], "a_x1": [2], "b_x2": [3], "a_y1": [4]})
    >>> names = ["condition", "group", "number"]
    >>> pat = "(.*)_(.)(.*)"
    >>> pivot_longer(df2, _["a_x1":"a_y1"], names_to = names, names_pattern = pat)
       id condition group number  value
    0   1         a     x      1      2
    0   1         b     x      2      3
    0   1         a     y      1      4

    >>> names = ["x1", "x2", "y1", "y2"]
    >>> wide = pd.DataFrame({
    ...    "x1": [1, 11], "x2": [2, 22], "y1": [3, 33], "y2": [4, 44]
    ... })
    >>> pivot_longer(wide, _[:], names_to = [".value", "set"], names_pattern = "(.)(.)")
      set   x   y
    0   1   1   3
    0   2   2   4
    1   1  11  33
    1   2  22  44

    """

    df_spec = build_longer_spec(
        __data,
        *cols,
        names_to=names_to,
        values_to=values_to,
        names_prefix=names_prefix,
        names_sep=names_sep,
        names_pattern=names_pattern,
        names_ptypes=names_ptypes,
    )

    return pivot_longer_spec(
        __data,
        df_spec,
        names_repair,
        values_drop_na,
        values_ptypes, 
        values_transform
    )


@pivot_longer.register(DataFrameGroupBy)
def _pivot_longer_gdf(__data, *args, **kwargs):
    # TODO: consolidate all verbs that punt to DataFrame version (#118)
    prior_groups = [el.name for el in __data.grouper.groupings]

    df = __data.obj
    res = pivot_longer(df, *args, **kwargs)

    missing_groups = set(prior_groups) - set(res.columns)
    if missing_groups:
        raise ValueError(
                "When using pivot_longer on grouped data, the result must contain "
                "original grouping columns. Missing group columns: %s" %missing_groups
                )

    return res.groupby(prior_groups)


@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def build_longer_spec(
    __data,
    *cols,
    names_to="name",
    values_to="value",
    names_prefix: "str | None"=None,
    names_sep=None,
    names_pattern=None,
    names_ptypes=None,
    names_transform: "dict[str, Callable] | None"=None
):
    if names_sep is not None and names_pattern is not None:
        raise ValueError("You may only use either `names_sep` or "
                         "`names_pattern`.")

    if isinstance(names_to, str):
        names_to = (names_to,)

    # select id columns and measure data --------------------------------------

    var_list = var_create(*cols)
    od = var_select(__data.columns, *var_list)

    value_vars = list(od)

    if not value_vars:
        raise ValueError(
            "Please select at least 1 column of values in pivot_longer.\n\n"
            "E.g. pivot_longer(data, _.some_col, _.another_col, ...)"
        )

    # note that this will include repeats in the data (e.g. two columns named "a")
    wide_values = __data.loc[:,value_vars]
    wide_cols = list(wide_values.columns)

    # strip prefix ------------------------------------------------------------
    if names_prefix is None:
        names = wide_cols
    elif isinstance(names_prefix, str):
        names = [re.sub(f"^{names_prefix}", "", name) for name in wide_cols]
    else:
        raise TypeError("names_prefix must be a string or None.")

    # start spec and split name into parts ------------------------------------
    # note that we set .name to be the names with names_prefix removed, do all
    # of the part splitting off that name, then set .name to the original values
    # at the very end.
    df_spec = pd.DataFrame({".name": names, ".value": values_to})

    if names_sep:
        df_spec = separate(df_spec, ".name", names_to, names_sep, remove=False)
    elif names_pattern:
        df_spec = extract(df_spec, ".name", names_to, names_pattern, remove=False)
    else:
        if len(names_to) == 1:
            df_spec = df_spec.assign(**{names_to[0]: df_spec[".name"]})
        else:
            raise TypeError(
                "pivot_longer either needs names_to to be string, or to receive "
                "names_sep or names_pattern arguments."
            )

    # setting names back to original
    df_spec[".name"] = wide_cols

    # transform columns -------------------------------------------------------
    if names_transform:
        _cols = list(df_spec.columns[2:])
        transforms = check_dict_of_functions(names_transform, _cols, "names_transform")

        for col_name, func in transforms.items():
            df_spec[col_name] = func(df_spec[col_name])

    return df_spec
    

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def pivot_longer_spec(
    __data,
    spec,
    names_repair: Optional[str] = "check_unique",
    values_drop_na: bool = False,
    values_ptypes = None,
    values_transform = None
):

    column_index = spec_to_multiindex(spec)

    wide_values = __data.loc[:,spec[".name"].unique()]
    wide_ids = __data.loc[:,~__data.columns.isin(wide_values.columns)]


    # reshape to long ---------------------------------------------------------
    inner_levels_na = pd.isna(column_index.names[1:])
    indx_to_drop = np.where(inner_levels_na)[0]

    if len(column_index.levels[0]) == 1:
        # simple case: only creating a single value column. in this case we use pd.melt,
        # since it can handle duplicate single and multi-index columns.
        _value_name = column_index.levels[0][0]

        wide_values.columns = column_index.droplevel(0)

        long_values = pd.melt(
            wide_values,
            value_vars=None,
            value_name=_value_name,
            ignore_index=False
        ).sort_index(level=-1).pipe(_drop_cols_by_position, indx_to_drop)

    else:
        # complex case: multiple value columns. Note that pandas throws an error if the
        # columns being unstacked contain duplicate column parts. E.g. if you split
        # duplicate columns x_1_1 and x_1_1 on "_".
        # (this behavior is fairly funky in dplyr, so probably okay to error)
        n_orig_levels = wide_values.index.nlevels

        if column_index.nlevels > 1:
            inner_levels = np.array(range(1, column_index.nlevels))
        else:
            # edge case, where only _values exists and is defined from matching
            # patterns in the columns. we can stack this only multiindex level, and it
            # will still set the _values as columns.
            inner_levels = np.array([0])
            indx_to_drop = np.array([0])

        wide_values.columns = column_index

        # note that levels named NA should not be in the final result, but are
        # needed for stacking.
        long_values_almost = (wide_values
            .stack(inner_levels.tolist())
            .droplevel(list(n_orig_levels + indx_to_drop))
        )

        # note: this is necessary for pandas <1.3 backwards compat. our column
        # index is a categorical, so reset_index is seen as trying to add a category
        # that doesn't exist to it... :/
        long_values_almost.columns = list(long_values_almost.columns)

        # once we drop earlier pandas versions, can cut out piece above
        long_values = (long_values_almost
            .reset_index(inner_levels[~inner_levels_na].tolist())
        )
    
    if values_drop_na:
        value_column_names = list(column_index.levels[0])
        long_values = long_values.dropna(subset=value_column_names, how="all")

    # transform values --------------------------------------------------------

    # TODO: names_transform for names, values_transform for values (including .value in names_to)

    transformed = long_values

    if values_transform:
        value_names = list(spec[".value"].unique())
        transforms = check_dict_of_functions(
            values_transform,
            value_names,
            "values_transform"
        )

        for col_name, f_transform in transforms.items():
            # TODO: error handling -- this won't work for dictionaries
            # list needs special handling, as it can only be applied to iterables,
            # not integers.
            long_values[col_name] = f_transform(long_values[col_name])

    # merge in id variables ---------------------------------------------------

    if wide_ids.shape[1] == 0:
        # no id columns, just return long values
        output_df = transformed
    else:
        output_df = pd.merge(wide_ids, transformed, left_index=True, right_index=True)
    
    # note that the pandas merge above has to add a suffix to duplicate columns,
    # so we need to use the original column names when repairing/validating.
    repaired_names = vec_as_names([*wide_ids.columns, *transformed.columns], repair=names_repair)
    output_df.columns = repaired_names

    return output_df





from siuba.siu import symbolic_dispatch
from siuba.sql.translate import SqlColumn, SqlColumnAgg
from sqlalchemy.sql import func as fn
from sqlalchemy import types as sa_types
from sqlalchemy import sql

@symbolic_dispatch(cls = SqlColumn)
def date_trunc(_, col, period):
    return fn.date_trunc(period, col)


# TODO: MC-NOTE: make a date_trunc generic?
# for mysql, could implement as https://stackoverflow.com/a/32955740

@symbolic_dispatch(cls = SqlColumn)
def sql_func_last_day_in_period(codata, col, period):
    return date_trunc(codata, col, period) + sql.text("INTERVAL '1 %s' - INTERVAL '1 day'" % period)


# TODO: RENAME TO GEN
# MYSQL: sql_is_date_offset(period, is_start=True)
@symbolic_dispatch(cls = SqlColumn)
def sql_is_first_day_of(codata, col, period):
    return date_trunc(codata, col, "day") == date_trunc(codata, col, period)


# TODO: how to handle this function generator?
# RENAME TO GEN
@symbolic_dispatch(cls = SqlColumn)
def sql_is_last_day_of(codata, col, period):
    last_day = sql_func_last_day_in_period(codata, col, period)
    return date_trunc(codata, col, 'day') == last_day


# MYSQL FUNCS ===

# TODO: easy to register
# def sql_func_extract_dow_monday(_, col):
#     # MYSQL: sunday starts, equals 1 (an int)
#     # pandas: monday starts, equals 0 (also an int)
# 
#     raw_dow = fn.dayofweek(col)
# 
#     # monday is 2 in MYSQL, so use monday + 5 % 7
#     return (raw_dow + 5) % 7

def sql_is_date_offset(period, is_start = True):

    # will check against one day in the past for is_start, v.v. otherwise
    fn_add = fn.date_sub if is_start else fn.date_add

    def f(_, col):
        get_period = lambda col: fn.extract(period, col)
        src_per = get_period(col)
        incr_per = get_period(fn_add(col, sql.text("INTERVAL 1 DAY")))

        return src_per != incr_per

    return f



__all__ = [
        "mtcars",
        "cars",
        "penguins",
        "penguins_raw",
        "cars_sql",
        "band_members",
        "band_instruments",
        "band_instruments2",
        "fish_encounters",
        "us_rent_income",
        "warpbreaks",
        ]

def __dir__():
    return __all__


def _load_data_csv_gz(name):
    import pandas as pd
    import pkg_resources

    fname = pkg_resources.resource_filename("siuba.data", f"{name}.csv.gz")
    return pd.read_csv(fname)


def _load_data_csv(name):
    import pandas as pd
    import pkg_resources

    fname = pkg_resources.resource_filename("siuba.data", f"{name}.csv")
    return pd.read_csv(fname)


def _load_data_cars_sql():
    import siuba.sql.utils as _sql_utils
    from siuba.sql import LazyTbl as _LazyTbl
    cars_sql = _LazyTbl(
            _sql_utils.mock_sqlalchemy_engine("postgresql"),
            "cars",
            ["cyl", "mpg", "hp"]
            )
    
    return cars_sql


def __getattr__(name):
    if name not in __all__:
        raise AttributeError(f"No dataset named: {name}")

    if name == "cars":
        return _load_data_csv_gz("mtcars")[["cyl", "mpg", "hp"]]

    elif name == "cars_sql":
        return _load_data_cars_sql() 

    elif name in {"band_members", "band_instruments", "band_instruments2"}:
        return _load_data_csv(name)

    return _load_data_csv_gz(name)

# cars_sql --------------------------------------------------------------------



import importlib

from contextlib import contextmanager

try:
    # once we drop sqlalchemy 1.2, can use create_mock_engine function
    from sqlalchemy.engine.mock import MockConnection
except ImportError:
    from sqlalchemy.engine.strategies import MockEngineStrategy
    MockConnection = MockEngineStrategy.MockConnection


def get_dialect_translator(name):
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    return mod.translator

def get_dialect_funcs(name):
    #dialect = engine.dialect.name
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    return mod.funcs

def get_sql_classes(name):
    mod = importlib.import_module('siuba.sql.dialects.{}'.format(name))
    win_name = name.title() + "Column"
    agg_name = name.title() + "ColumnAgg"
    return {
            'window': getattr(mod, win_name),
            'aggregate': getattr(mod, agg_name)
            }


def mock_sqlalchemy_engine(dialect):
    """
    Create a sqlalchemy.engine.Engine without it connecting to a database.

    Examples
    --------

    ::
        from siuba.sql import LazyTbl
        from siuba import _, mutate, show_query

        engine = mock_sqlalchemy_engine('postgresql')
        tbl = LazyTbl(engine, 'some_table', ['x'])

        query = mutate(tbl, y = _.x + _.x)
        show_query(query)

    """

    from sqlalchemy.engine import Engine
    from sqlalchemy.dialects import registry
    from types import ModuleType

    #  TODO: can be removed once v1.3.18 support dropped
    from sqlalchemy.engine.url import URL


    dialect_cls = registry.load(dialect)

    # there is probably a better way to do this, but for some reason duckdb
    # returns a module, rather than the dialect class itself. By convention,
    # dialect modules expose a variable named dialect, so we grab that.
    if isinstance(dialect_cls, ModuleType):
        dialect_cls = dialect_cls.dialect
    
    conn = MockConnection(dialect_cls(), lambda *args, **kwargs: None)

    # set a url on it, so that LazyTbl can read the backend name.
    if is_sqla_12() or is_sqla_13():
        url = URL(drivername=dialect)
    else:
        url = URL.create(drivername=dialect)

    conn.url = url

    return conn


# Temporary fix for pandas bug (https://github.com/pandas-dev/pandas/issues/35484)
from pandas.io import sql as _pd_sql

class _FixedSqlDatabase(_pd_sql.SQLDatabase):
    def execute(self, *args, **kwargs):
        if hasattr(self, "connectable"):
            return self.connectable.execute(*args, **kwargs)
        
        return self.con.execute(*args, **kwargs)


# Detect duckdb for temporary workarounds -------------------------------------

def _is_dialect_duckdb(engine):
    return engine.url.get_backend_name() == "duckdb"

# Backwards compatibility for sqlalchemy 1.3 ----------------------------------

import re
import sqlalchemy

RE_VERSION=r"(?P<major>\d+)\.(?P<minor>\d+).(?P<patch>\d+)"  
SQLA_VERSION=tuple(map(int, re.match(RE_VERSION, sqlalchemy.__version__).groups()))

def is_sqla_12():
    return SQLA_VERSION[:-1] == (1, 2)

def is_sqla_13():
    return SQLA_VERSION[:-1] == (1, 3)

def is_sqla_14():
    return SQLA_VERSION[:-1] == (1, 4)


def _sql_select(columns, *args, **kwargs):
    from sqlalchemy import sql
    if is_sqla_12() or is_sqla_13():
        # use old syntax, where columns are passed as a list
        return sql.select(columns, *args, **kwargs)

    return sql.select(*columns, *args, **kwargs)


def _sql_column_collection(columns):
    # This function largely handles the removal of ImmutableColumnCollection in
    # sqlalchemy, in favor of ColumnCollection being immutable.

    data = {col.key: col for col in columns}

    if is_sqla_12() or is_sqla_13():
        from sqlalchemy.sql.base import ImmutableColumnCollection

        return ImmutableColumnCollection(data, columns)

    elif is_sqla_14():
        from sqlalchemy.sql.base import ColumnCollection

        return ColumnCollection(list(data.items())).as_immutable()

    else:
        from sqlalchemy.sql.base import ColumnCollection

        return ColumnCollection(list(data.items()))


def _sql_add_columns(select, columns):
    if is_sqla_12() or is_sqla_13():
        for column in columns:
            select = select.column(column)
        return select

    return select.add_columns(*columns)


def _sql_with_only_columns(select, columns):
    if is_sqla_12() or is_sqla_13():
        out = select.with_only_columns(columns)
    else:
        out = select.with_only_columns(*columns)

    # ensure removing all columns doesn't remove from clause table reference
    for _from in select.froms:
        out = out.select_from(_from)

    return out


def _sql_case(whens, **kwargs):
    from sqlalchemy import sql
    if is_sqla_12() or is_sqla_13():
        return sql.case(whens, **kwargs)

    return sql.case(*whens, **kwargs)


# Simplify option in show_query -----------------------------------------------

def _sql_refresh(el):
    f_refresh = getattr(
        el,
        "_reset_memoizations",
        getattr(el, "_reset_exported", None)
    )

    f_refresh()


def _sql_simplify_select(select):
    from sqlalchemy import sql
    from sqlalchemy.sql.visitors import traverse, cloned_traverse

    def simplify_sel(sel):

        if isinstance(sel, sql.Select) and len(sel.froms) == 1:
            child = sel.froms[0]

            # technically should be an ordered set
            crnt_cols = set(sel.inner_columns)
            child_cols = set(child.columns)

            if len(child_cols - crnt_cols) == 0:
                remaining = list(crnt_cols - child_cols)

                star = sql.text("*")
                star._from_objects = (child,)

                sel._raw_columns = [star, *remaining]
                _sql_refresh(sel)

    # TODO: find simpler way to clone an element. We cannot use the visitors
    # argument of cloned_traverse, since it visits the inner-most element first.
    clone_el = select._clone()

    # modify in-place
    traverse(clone_el, {}, {"select": simplify_sel})

    return clone_el



def lift_inner_cols(tbl):
    cols = list(tbl.inner_columns)

    return _sql_column_collection(cols)




from .calls import (
    Call,
    MetaArg,
    FuncArg,
    Lazy,
    BinaryOp,
    _SliceOpIndex,
    DictCall,
    FormulaArg,
    FormulaContext,
    str_to_getitem_call
)
from .symbolic import Symbolic, strip_symbolic, create_sym_call, explain
from .visitors import CallTreeLocal, CallVisitor, FunctionLookupBound, FunctionLookupError, ExecutionValidatorVisitor
from .dispatchers import symbolic_dispatch, singledispatch2, pipe_no_args, Pipeable, pipe, call

Lam = Lazy

_ = Symbolic()
Fx = Symbolic(FormulaArg("Fx"))





import pytest
from .helpers import assert_equal_query, data_frame

from siuba import _, head, group_by, arrange


DATA = data_frame(a = [1,2,3,4,5,6], g = ["x", "y"] * 3)


@pytest.fixture
def df(backend):
    # Note that we use arrange the data, so that cloud backends that do not
    # guarantee order without ORDER BY will return the right thing
    return backend.load_df(DATA) >> arrange(_.a)


def test_head_default(df):
    assert_equal_query(
            df,
            head(),
            DATA.iloc[:5,:]
    )


def test_head_n_arg(df):
    assert_equal_query(
            df,
            head(2),
            DATA.iloc[:2, :]
    )


def test_head_grouped(df):
    assert_equal_query(
            df,
            group_by(_.g) >> head(),
            DATA.iloc[:5, :]
    )


def test_head_grouped_n_arg(df):
    assert_equal_query(
            df,
            group_by(_.g) >> head(2),
            DATA.iloc[:2, :]
    )



from IPython import get_ipython
from IPython.core.history import HistoryAccessor
from IPython.core.completer import cursor_to_position, _FakeJediCompletion
from typing import Iterable, Any
from functools import wraps, partial
from contextlib import contextmanager

from siuba.siu import Symbolic


def order_results_cols_first(completions: "iter[Any]", df) -> "list[Any]":
    # TODO: annotate this as a list of jedi completions objs
    cols = set(df.columns)
    return sorted(completions, key = lambda x: 0 if x.name in cols else 1)


def wrap_jedi_matches(_jedi_matches):
    @wraps(_jedi_matches)
    def wrapper(self, *args, **kwargs):
        orig_namespace = self.namespace

        # find DataFrame and inject into namespace
        df_name = _find_df_in_history(self.shell)

        # guard to fallback to default behavior ----
        if df_name is None or not isinstance(self.shell.user_ns.get("_"), Symbolic):
            return _jedi_matches(*args, **kwargs)

        # insert dataframe and autocomplete ----
        df = self.namespace[df_name]
        self.namespace = {**self.namespace, "_": df}

        try:
            completions = _jedi_matches(*args, **kwargs)
        finally:
            self.namespace = orig_namespace

        self._all_completions = list(completions)

        return order_results_cols_first(self._all_completions, df)

        #return completions

    return wrapper


def _match_df_name(dfs, commands):
    if not dfs:
        return []

    for command in reversed(commands):
        exact_match = [df for df in dfs if df == command]
        if exact_match:
            return exact_match[0]

        method_match = [df for df in dfs if df + "." in command]
        if method_match:
            return method_match[0]

        assign_match = [df for df in dfs if command.startswith(df + " = ")]
        if assign_match:
            return assign_match[0]

        import_match = [df for df in dfs if "import " + df in command]
        if import_match:
            return import_match[0]

        in_expression = [df for df in dfs if df + "," in command or df + ")" in command]
        if in_expression:
            return in_expression[0]

        any_match = [df for df in dfs if df in command]
        if any_match:
            return any_match[0]

    # return one of the dataframes
    return dfs[0]


def _find_df_in_history(shell):
    dfs = shell.run_line_magic("who_ls", "DataFrame")

    history = HistoryAccessor()
    commands = [command for _, _, command in history.get_tail(include_latest=True)]

    return _match_df_name(dfs, commands)


def siuba_jedi_override(shell):
    # wrap the bound method _jedi_matches. Note that Completer is actually an instance
    self = shell.Completer
    wrapped = wrap_jedi_matches(self._jedi_matches)
    shell.Completer._jedi_matches = partial(wrapped, self)


shell = get_ipython()

if shell is not None:
    siuba_jedi_override(shell)



from siuba.siu import Symbolic, strip_symbolic
from siuba.ops.support import spec
from .helpers import (
    data_frame,
    assert_equal_query,
    backend_pandas,
    pd_version,
    SqlBackend,
    PandasBackend
)
import pytest
# TODO: dot, corr, cov

from siuba import filter, mutate, summarize, group_by, arrange
from pandas.testing import assert_frame_equal, assert_series_equal
import numpy as np
import pandas as pd
import pkg_resources


def get_action_kind(spec_entry):
    return spec_entry["kind"]

def filter_entry(spec, f):
    out = []
    for k, v in spec.items():
        raw_kind = v.get("kind")
        kind = raw_kind.title() if raw_kind is not None else "Unknown"
        status = v["backends"].get("pandas", {}).get("support", "Supported").title()

        if f(kind, status):
            out.append(k)

    return out

SPEC_IMPLEMENTED = filter_entry(spec, lambda k, s: s == "Supported")
SPEC_NOTIMPLEMENTED = filter_entry(spec, lambda k, s: s != "Supported")
SPEC_AGG = filter_entry(spec, lambda k, s: k in {"Agg"} and s == "Supported")

REMOVED_IN_PANDAS_V2 = {
    "week",
    "weekofyear",
    "mad",
}


_ = Symbolic()

@pytest.fixture(params = tuple(SPEC_IMPLEMENTED))
def entry(request):
    # NOTE: the sole purpose of putting in a fixture is so pytest line output
    #       is very easy to read. (e.g. pytest -v --tb=line)
    key = request.param
    yield spec[key]

@pytest.fixture(params = tuple(SPEC_AGG))
def agg_entry(request):
    key = request.param
    yield spec[key]

@pytest.fixture(params = tuple(SPEC_NOTIMPLEMENTED))
def notimpl_entry(request):
    key  = request.param
    yield spec[key]

def assert_src_array_equal(src, dst):
    if isinstance(src, np.ndarray):
        assert np.array_equal(src, dst)
    elif isinstance(src, pd.DataFrame):
        assert_frame_equal(src, dst)
    elif isinstance(src, pd.Series):
        assert_series_equal(src, dst, check_names = False)
    else:
        assert src == dst


def skip_if_removed(entry):
    if pd_version >= (2, 0, 0) and entry["name"] in REMOVED_IN_PANDAS_V2:
        pytest.skip()
    
# Data ========================================================================
data_dt = data_frame(
    id = [1, 2, 3, 4],
    g = ['a', 'a', 'b', 'b'],
    x = pd.to_datetime(["2019-01-01 01:01:01", "2020-04-08 02:02:02", "2021-07-15 03:03:03", "2022-10-22 04:04:04"])
    )

data_str = data_frame(
    id = [1, 2, 3, 4],
    g = ['a', 'a', 'b', 'b'],
    x = ['abc', 'cde', 'fg', 'h']
    )

data_bool = data_frame(
    id = [1, 2, 3, 4],
    g = ['a', 'a', 'b', 'b'],
    x = [True, False, True, False],
    y = [True, True, False, False]
        )

data_cat = data_frame(
    id = [1, 2, 3, 4],
    g = ['a', 'a', 'b', 'b'],
    x = pd.Categorical(['abc', 'cde', 'fg', 'h'])
    )

data_sparse = data_frame(
    id = [1, 2, 3, 4],
    g = ['a', 'a', 'b', 'b'],
    x = pd.Series([1, 2, 3, 4], dtype = "Sparse")
    )

data_default = data_frame(
    id = [1, 2, 3, 4, 5, 6],
    g = ['a', 'a', 'a', 'b', 'b', 'b'],
    x = [10, 11, 11, 13, 13, 13],
    y = [1,2,3,4,5,6]
    )

DATA = data = {
    'dt': data_dt,
    'str': data_str,
    None: data_default,
    'bool': data_bool,
    'cat': data_cat,
    'sparse': data_sparse

}

def get_spec_no_mutate(entry, backend):
    return "no_mutate" in entry['backends'].get(backend.name, {}).get('flags', [])

def get_spec_no_summarize(entry, backend):
    return "no_aggregate" in entry['backends'].get(backend.name, {}).get('flags', [])
    
def get_spec_backend_is_supported(entry, backend):
    return entry['backends'].get(backend.name, {}).get('is_supported')

def get_spec_sql_type(entry, backend):
    return entry['backends'].get(backend.name, {}).get('result_type')

def get_data(entry, data, backend = None):

    if backend:
        req_bool = entry["backends"].get(backend.name, {}).get('input_type') == 'bool'
    else:
        req_bool = False

    return data['bool'] if req_bool else data[entry['accessor']]

def skip_no_mutate(entry, backend):
    if "no_mutate" in entry['backends'].get(backend.name, {}).get('flags', []):
        pytest.skip("No support for mutate in backend: %s" %backend.name)

def skip_no_summarize(entry, backend):
    if "no_aggregate" in entry['backends'].get(backend.name, {}).get('flags', []):
        pytest.skip("No support for summarize in backend: %s" %backend.name)

def do_test_missing_implementation(entry, backend):
    # Check whether test should xfail, skip, or -------------------------------
    supported = get_spec_backend_is_supported(entry, backend)

    if not supported:
        pytest.skip("Spec'd failure")

    #if get_spec_no_mutate(entry, backend):
    #    pytest.skip("Spec'd failure")

    ## case: Needs to be implemented
    ## TODO(table): uses xfail
    #if backend_status == "todo":
    #    pytest.xfail("TODO: implement this translation")
    #
    ## case: Can't be used in a mutate (e.g. a SQL ordered set aggregate function)
    ## TODO(table): no_mutate

    ## case: won't be implemented
    #if get_spec_backend_status(entry, backend.name) == "wontdo":
    #    pytest.skip()


def get_df_expr(entry):
    str_expr = str(entry['expr_frame'])
    #call_expr = strip_symbolic(eval(str_expr, {'_': _}))
    return str_expr, entry["expr_frame"]

    return str_expr, call_expr

def cast_result_type(entry, backend, ser):
    sql_type = get_spec_sql_type(entry, backend)
    if isinstance(backend, SqlBackend) and sql_type == 'float':
        return ser.astype('float')
    elif isinstance(backend, SqlBackend) and sql_type == 'int':
        return ser.astype('int')
    
    return ser

# Tests =======================================================================


def test_series_against_call(entry):
    # TODO: this test originally was evaluating string representations created
    # by calls. I've changed it to just executing the calls directly.
    # not sure the original intent of the test?

    skip_if_removed(entry)

    if entry['kind'] == "window":
        pytest.skip()

    df = data[entry['accessor']]
    # TODO: once reading from yaml, no need to repr
    str_expr = str(entry['expr_series'])

    call_expr = entry['expr_series']#  strip_symbolic(eval(str_expr, {'_': _}))
    res = call_expr(df.x)

    dst = call_expr(df.x)# eval(str_expr, {'_': df.x})
    
    assert res.__class__ is dst.__class__
    assert_src_array_equal(res, dst)


def test_frame_expr(entry):
    skip_if_removed(entry)

    # TODO: remove this test, not checking anything new
    df = data[entry['accessor']]
    # TODO: once reading from yaml, no need to repr
    str_expr = str(entry['expr_frame'])

    call_expr = entry['expr_frame']
    res = call_expr(df)

    dst = call_expr(df)
    
    assert res.__class__ is dst.__class__
    assert_src_array_equal(res, dst)


def test_pandas_grouped_frame_fast_not_implemented(notimpl_entry):
    from siuba.experimental.pd_groups.dialect import fast_mutate
    gdf = data[notimpl_entry['accessor']].groupby('g')

    # TODO: once reading from yaml, no need to repr
    call_expr = notimpl_entry['expr_frame']

    with pytest.warns(UserWarning):
        try:
            # not implemented functions are punted to apply, and
            # not guaranteed to work (e.g. many lengthen arrays, etc..)
            res = fast_mutate(gdf, result = call_expr)
        except:
            pass
    


#@backend_pandas
def test_frame_mutate(skip_backend, backend, entry):
    do_test_missing_implementation(entry, backend)
    skip_no_mutate(entry, backend)

    skip_if_removed(entry)

    # Prepare input data ------------------------------------------------------
    # case: inputs must be boolean
    crnt_data = get_data(entry, DATA, backend)
    df = backend.load_cached_df(crnt_data)

    # Execute mutate ----------------------------------------------------------
    str_expr, call_expr = get_df_expr(entry)

    # Run test for equality w/ ungrouped pandas ----
    dst = crnt_data.assign(result = call_expr(crnt_data))
    dst['result'] = cast_result_type(entry, backend, dst['result'])

    result_type = get_spec_sql_type(entry, backend)

    # relax dtype checking, since database queries sometimes return
    # int64 instead of int32, etc..
    kwargs = {"check_dtype": False, "atol": 1e-03}

    assert_equal_query(
            df,
            arrange(_.id) >> mutate(result = call_expr),
            dst,
            **kwargs
            )

    # Run test for equality w/ grouped pandas ----
    g_dst = crnt_data.groupby('g').apply(lambda d: d.assign(result = call_expr)).reset_index(drop = True)
    g_dst['result'] = cast_result_type(entry, backend, g_dst['result'])
    assert_equal_query(
            df,
            arrange(_.id) >> group_by(_.g) >> mutate(result = call_expr),
            g_dst,
            **kwargs
            )


def test_pandas_grouped_frame_fast_mutate(entry):
    from siuba.experimental.pd_groups.dialect import fast_mutate, DataFrameGroupBy

    skip_if_removed(entry)
    
    gdf = get_data(entry, DATA).groupby('g')

    # Execute mutate ----------------------------------------------------------
    str_expr, call_expr = get_df_expr(entry)

    res = fast_mutate(gdf, result = call_expr)
    dst = mutate(gdf, result = call_expr)

    # TODO: apply mark to skip failing tests, rather than casting?
    # in pandas 1.2, grouped agg returns int, ungrouped agg returns float
    # in pandas 1.3, grouped agg returns float, same as ungrouped agg
    # (the difference is because the grouped agg in 1.2 did not use cython,
    # and tries casting back to the original column dtype)
    res_obj = res.obj
    if str_expr == '_.x.median()':
        res_obj['result'] = res_obj['result'].astype(float)

    assert isinstance(dst, DataFrameGroupBy)
    assert_frame_equal(res_obj, dst.obj)


def test_frame_summarize(skip_backend, backend, agg_entry):
    skip_if_removed(agg_entry)

    entry = agg_entry

    do_test_missing_implementation(entry, backend)
    skip_no_summarize(entry, backend)

    # Prepare input data ------------------------------------------------------
    # case: inputs must be boolean
    crnt_data = get_data(entry, DATA, backend)
    df = backend.load_cached_df(crnt_data)

    # Execute mutate ----------------------------------------------------------
    str_expr, call_expr = get_df_expr(entry)

    dst = data_frame(result = call_expr(crnt_data))

    # Process output ----------------------------------------------------------
    # case: output is of a different type than w/ pandas
    dst['result'] = cast_result_type(entry, backend, dst['result'])

    result_type = get_spec_sql_type(entry, backend)
    if result_type == "variable":
        kwargs = {"check_dtype": False, "atol": 1e-03}
    else:
        kwargs = {}

    # Run test for equality w/ pandas ----
    # otherwise, verify returns same result as mutate
    assert_equal_query(
            df,
            summarize(result = call_expr),
            dst,
            **kwargs
            )

    dst_g = crnt_data.groupby('g').apply(call_expr).reset_index().rename(columns = {0: 'result'})
    dst_g["result"] = cast_result_type(entry, backend, dst_g['result'])
    assert_equal_query(
            df,
            group_by(_.g) >> summarize(result = call_expr),
            dst_g,
            **kwargs
            )


def test_pandas_grouped_frame_fast_summarize(agg_entry):
    from siuba.experimental.pd_groups.dialect import fast_summarize, DataFrameGroupBy

    skip_if_removed(agg_entry)

    gdf = get_data(agg_entry, DATA).groupby('g')

    # Execute summarize ----------------------------------------------------------
    str_expr, call_expr = get_df_expr(agg_entry)

    res = fast_summarize(gdf, result = call_expr)
    dst = summarize(gdf, result = call_expr)

    # TODO: apply mark to skip failing tests, rather than downcast
    # pandas grouped aggs, when not using cython, _try_cast back to original type
    # but since summarize uses apply, it doesn't :/. Currently only affects median func.
    if str_expr == '_.x.median()':
        res['result'] = res['result'].astype(float)

    assert_frame_equal(res, dst)



# Edge Cases ==================================================================

@pytest.mark.postgresql
def test_frame_set_aggregates_postgresql():
    # TODO: probably shouldn't be creating backend here
    backend = SqlBackend("postgresql")
    dfs = backend.load_cached_df(data[None])
    
    expr = _.x.quantile(.75)
    assert_equal_query(
            dfs,
            group_by(_.g) >> summarize(result = expr),
            data_frame(g = ['a', 'b'], result = [11., 13.])
            )





"""
Implements LazyTbl to represent tables of SQL data, and registers it on verbs.

This module is responsible for the handling of the "table" side of things, while
translate.py handles translating column operations.


"""

import warnings

from .translate import CustomOverClause, ColumnCollection
from .utils import (
    get_dialect_translator,
    _sql_column_collection,
    _sql_add_columns,
)

from sqlalchemy import sql
import sqlalchemy
from siuba.siu import FunctionLookupError
from functools import singledispatch

# Helpers ---------------------------------------------------------------------

class SqlFunctionLookupError(FunctionLookupError): pass

class WindowReplacer:

    @staticmethod
    def _get_unique_name(prefix, columns):
        column_names = set(columns.keys())

        i = 1
        name = prefix + str(i)
        while name in column_names:
            i += 1
            name = prefix + str(i)


        return name

    @staticmethod
    def _get_over_clauses(clause):
        windows = []
        append_win = lambda col: windows.append(col)

        sql.util.visitors.traverse(clause, {}, {"over": append_win})

        return windows


class SqlLabelReplacer:
    """Create a visitor to replace source labels with destination.

    Note that this is meant to be used with sqlalchemy visitors.
    """

    def __init__(self, src_columns, dst_columns):
        self.src_columns = src_columns
        self.src_labels = set([x for x in src_columns if isinstance(x, sql.elements.Label)])
        self.dst_columns = dst_columns
        self.applied = False

    def __call__(self, clause):
        return sql.util.visitors.replacement_traverse(clause, {}, self.visit)
    
    def visit(self, el):
        from sqlalchemy.sql.elements import ColumnClause, Label, ClauseElement, TypeClause
        from sqlalchemy.sql.schema import Column

        if isinstance(el, TypeClause):
            # TODO: for some reason this type throws an error if unguarded
            return None

        if isinstance(el, ClauseElement):
            if el in self.src_labels:
                self.applied = True
                return self.dst_columns[el.name]
            elif el in self.src_columns:
                return self.dst_columns[el.name]

            #elif isinstance(el, ColumnClause) and not isinstance(el, Column):
            #    # Raw SQL, which will need a subquery, but not substitution
            #    if el.key != "*":
            #        self.applied = True
        
        return None
            

def track_call_windows(call, columns, group_by, order_by, window_cte = None):
    col_expr = call(columns)

    crnt_group_by = sql.elements.ClauseList(
            *[columns[name] for name in group_by]
            )

    crnt_order_by = sql.elements.ClauseList(*order_by)

    return replace_call_windows(col_expr, crnt_group_by, crnt_order_by, window_cte)



@singledispatch
def replace_call_windows(col_expr, group_by, order_by, window_cte = None):
    raise TypeError(str(type(col_expr)))


@replace_call_windows.register(ColumnCollection)
def _(col_expr, group_by, order_by, window_cte = None):
    all_over_clauses = []
    for col in col_expr:
        _, over_clauses, window_cte = replace_call_windows(
            col,
            group_by,
            order_by,
            window_cte
        )
        all_over_clauses.extend(over_clauses)

    return col_expr, all_over_clauses, window_cte


@replace_call_windows.register(sql.elements.ClauseElement)
def _(col_expr, group_by, order_by, window_cte = None):

    over_clauses = WindowReplacer._get_over_clauses(col_expr)

    for over in over_clauses:
        # TODO: shouldn't mutate these over clauses
        over.set_over(group_by, order_by)

    if len(over_clauses) and window_cte is not None:
        # custom name, or parameters like "%(...)s" may nest and break psycopg2
        # with columns you can set a key to fix this, but it doesn't seem to 
        # be an option with labels
        name = WindowReplacer._get_unique_name('win', lift_inner_cols(window_cte))
        label = col_expr.label(name)

        # put into CTE, and return its resulting column, so that subsequent
        # operations will refer to the window column on window_cte. Note that
        # the operations will use the actual column, so may need to use the
        # ClauseAdaptor to make it a reference to the label
        window_cte = _sql_add_columns(window_cte, [label])
        win_col = lift_inner_cols(window_cte).values()[-1]

        return win_col, over_clauses, window_cte
            
    return col_expr, over_clauses, window_cte

def get_single_from(sel):
    froms = sel.froms

    n_froms = len(froms)
    if n_froms != 1:
        raise ValueError(
            f"Expected a single table in the from clause, but found {n_froms}"
        )

    return froms[0]

def lift_inner_cols(tbl):
    cols = list(tbl.inner_columns)

    return _sql_column_collection(cols)

# Misc utilities --------------------------------------------------------------

def ordered_union(x, y):
    dx = {el: True for el in x}
    dy = {el: True for el in y}

    return tuple({**dx, **dy})


def _warn_missing(missing_groups):
    warnings.warn(f"Adding missing grouping variables: {missing_groups}")


# Table -----------------------------------------------------------------------

class LazyTbl:
    def __init__(
            self, source, tbl, columns = None,
            ops = None, group_by = tuple(), order_by = tuple(),
            translator = None
            ):
        """Create a representation of a SQL table.

        Args:
            source: a sqlalchemy.Engine or sqlalchemy.Connection instance.
            tbl: table of form 'schema_name.table_name', 'table_name', or sqlalchemy.Table.
            columns: if specified, a listlike of column names.

        Examples
        --------

        ::
            from sqlalchemy import create_engine
            from siuba.data import mtcars

            # create database and table
            engine = create_engine("sqlite:///:memory:")
            mtcars.to_sql('mtcars', engine)

            tbl_mtcars = LazyTbl(engine, 'mtcars')
            
        """
        
        # connection and dialect specific functions
        self.source = sqlalchemy.create_engine(source) if isinstance(source, str) else source

        # get dialect name
        dialect = self.source.dialect.name
        self.translator = get_dialect_translator(dialect)

        self.tbl = self._create_table(tbl, columns, self.source)

        # important states the query can be in (e.g. grouped)
        self.ops = [self.tbl] if ops is None else ops

        self.group_by = group_by
        self.order_by = order_by


    def append_op(self, op, **kwargs):
        cpy = self.copy(**kwargs)
        cpy.ops = cpy.ops + [op]
        return cpy

    def copy(self, **kwargs):
        return self.__class__(**{**self.__dict__, **kwargs})

    def shape_call(
            self,
            call, window = True, str_accessors = False,
            verb_name = None, arg_name = None,
            ):
        return self.translator.shape_call(call, window, str_accessors, verb_name, arg_name)

    def track_call_windows(self, call, columns = None, window_cte = None):
        """Returns tuple of (new column expression, list of window exprs)"""

        from .verbs.arrange import _eval_arrange_args

        columns = self.last_op.columns if columns is None else columns

        order_by = _eval_arrange_args(self, self.order_by, columns)

        return track_call_windows(call, columns, self.group_by, order_by, window_cte)

    def get_ordered_col_names(self):
        """Return columns from current select, with grouping columns first."""
        ungrouped = [k for k in self.last_op.columns.keys() if k not in self.group_by]
        return list(self.group_by) + ungrouped

    #def label_breaks_order_by(self, name):
    #    """Returns True if a new column label would break the order by vars."""

    #    # TODO: arrange currently allows literals, which breaks this. it seems
    #    #       better to only allow calls in arrange.
    #    order_by_vars = {c.op_vars(attr_calls=False) for c in self.order_by}




    @property
    def last_op(self) -> "sql.Table | sql.Select":
        last_op = self.ops[-1]

        if last_op is None:
            raise TypeError()

        return last_op

    @property
    def last_select(self):
        last_op = self.last_op
        if not isinstance(last_op, sql.selectable.SelectBase):
            return last_op.select()

        return last_op

    @staticmethod
    def _create_table(tbl, columns = None, source = None):
        """Return a sqlalchemy.Table, autoloading column info if needed. 

        Arguments:
            tbl: a sqlalchemy.Table or string of form 'table_name' or 'schema_name.table_name'.
            columns: a tuple of column names for the table. Overrides source argument.
            source: a sqlalchemy engine, used to autoload columns.

        """
        if isinstance(tbl, sql.selectable.FromClause):
            return tbl
        
        if not isinstance(tbl, str):
            raise ValueError("tbl must be a sqlalchemy Table or string, but was %s" %type(tbl))

        if columns is None and source is None:
            raise ValueError("One of columns or source must be specified")

        schema, table_name = tbl.split('.') if '.' in tbl else [None, tbl]

        columns = map(sqlalchemy.Column, columns) if columns is not None else tuple()

        # TODO: pybigquery uses schema to mean project_id, so we cannot use
        # siuba's classic breakdown "{schema}.{table_name}". Basically
        # pybigquery uses "{schema=project_id}.{dataset_dot_table_name}" in its internal
        # logic. An important side effect is that bigquery errors for
        # `dataset`.`table`, but not `dataset.table`.
        if source and source.dialect.name == "bigquery":
            table_name = tbl
            schema = None

        return sqlalchemy.Table(
                table_name,
                sqlalchemy.MetaData(),
                *columns,
                schema = schema,
                autoload_with = source if not columns else None
                )

    def _get_preview(self):
        # need to make prev op a cte, so we don't override any previous limit
        from siuba.dply.verbs import collect

        new_sel = self.last_select.limit(5)
        tbl_small = self.append_op(new_sel)
        return collect(tbl_small)

    def __repr__(self):
        template = (
                "# Source: lazy query\n"
                "# DB Conn: {}\n"
                "# Preview:\n{}\n"
                "# .. may have more rows"
                )

        return template.format(repr(self.source.engine), repr(self._get_preview()))

    def _repr_html_(self):
        template = (
                "<div>"
                "<pre>"
                "# Source: lazy query\n"
                "# DB Conn: {}\n"
                "# Preview:\n"
                "</pre>"
                "{}"
                "<p># .. may have more rows</p>"
                "</div>"
                )

        data = self._get_preview()

        # _repr_html_ can not exist or return None, to signify that repr should be used
        if not hasattr(data, '_repr_html_'):
            return None

        html_data = data._repr_html_()
        if html_data is None:
            return None

        return template.format(self.source.engine, html_data)


def _repr_grouped_df_html_(self):
    return "<div><p>(grouped data frame)</p>" + self._selected_obj._repr_html_() + "</div>"


sql_raw = sql.literal_column




import sqlalchemy as sqla
import uuid
import re

from siuba.sql import LazyTbl
from siuba.sql.utils import _is_dialect_duckdb
from siuba.dply.verbs import ungroup, collect
from siuba.siu import FunctionLookupError
from pandas.testing import assert_frame_equal
import pandas as pd
import os
import numpy as np

def data_frame(*args, _index = None, **kwargs):
    if len(args):
        raise NotImplementedError("all arguments to data_frame must be named")

    all_scalars = all(not np.ndim(v) for v in kwargs.values())

    if all_scalars:
        fixed = {k: [v] for k,v in kwargs.items()}
        return pd.DataFrame(fixed, index = _index)
    
    return pd.DataFrame(kwargs, index = _index)

BACKEND_CONFIG = {
        "postgresql": {
            "dialect": "postgresql",
            "driver": "",
            "dbname": ["SB_TEST_PGDATABASE", "postgres"],
            "port": ["SB_TEST_PGPORT", "5433"],
            "user": ["SB_TEST_PGUSER", "postgres"],
            "password": ["SB_TEST_PGPASSWORD", ""],
            "host": ["SB_TEST_PGHOST", "localhost"],
            "options": ""
            },
        "bigquery": {
            "dialect": "bigquery",
            # bigquery uses dbname for dataset
            "dbname": ["SB_TEST_BQDATABASE", "ci"],
            "port": "",
            "user": "",
            "password": "",
            "host": ["SB_TEST_BQPROJECT", "siuba-tests"],
            "options": ""
            },
        "snowflake": {
            "dialect": "snowflake",
            "driver": "",
            "dbname": ["SB_TEST_SNOWFLAKEDATABASE", "CI/TESTS"],
            "port": "",
            "user": ["SB_TEST_SNOWFLAKEUSER", "CI"],
            "password": ["SB_TEST_SNOWFLAKEPASSWORD", ""],
            "host": ["SB_TEST_SNOWFLAKEHOST", ""],
            "options": ["SB_TEST_SNOWFLAKEOPTIONS", "warehouse=COMPUTE_WH&role=CI_USER"]
            },
        "mysql": {
            "dialect": "mysql+pymysql",
            "dbname": "public",
            "port": ["SB_TEST_MYSQLPORT", 3306],
            "user": "root",
            "password": "",
            "host": "127.0.0.1",
            "options": ""
            },
        "sqlite": {
            "dialect": "sqlite",
            "driver": "",
            "dbname": ["SB_TEST_SQLITEDATABASE", ":memory:"],
            "port": "0",
            "user": "",
            "password": "",
            "host": "",
            "options": ""
            },
        "duckdb": {
            "dialect": "duckdb",
            "driver": "",
            "dbname": ":memory:",
            "port": "",
            "user": "",
            "password": "",
            "host": "",
            },
        }

class Backend:
    # TODO: use abstract base class
    kind = "pandas"

    def __init__(self, name):
        self.name = name

    def dispose(self):
        pass

    def load_df(self, df = None, **kwargs):
        if df is None and kwargs:
            df = pd.DataFrame(kwargs)
        elif df is not None and kwargs:
            raise ValueError("Cannot pass kwargs, and a DataFrame")

        return df

    def load_cached_df(self, df):
        return df

    def matches_test_qualifier(self, s):
        return self.name == s or self.kind == s

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, repr(self.name))

class PandasBackend(Backend):
    kind = "pandas"

class SqlBackend(Backend):
    kind = "sql"
    
    table_name_indx = 0

    # if there is a :, sqlalchemy tries to parse the port number.
    # since backends like bigquery do not specify a port, we'll insert it
    # later on the port value passed in.
    sa_conn_fmt = "{dialect}://{user}:{password}@{host}{port}/{dbname}?{options}"

    sa_conn_memory_fmt = "{dialect}:///{dbname}"

    def __init__(self, name):
        from urllib.parse import quote_plus

        cnfg = BACKEND_CONFIG[name]
        params = {k: os.environ.get(*v) if isinstance(v, (list)) else v for k,v in cnfg.items()}

        if params["port"]:
            params["port"] = ":%s" % params["port"]

        if params["password"]:
            params["password"] = quote_plus(params["password"])

        if params["dbname"] == ":memory:":
            sa_conn_uri = self.sa_conn_memory_fmt.format(**params)
        else:
            sa_conn_uri = self.sa_conn_fmt.format(**params)

        self.name = name
        self.engine = sqla.create_engine(sa_conn_uri)
        self.cache = {}

    def dispose(self):
        self.engine.dispose()

    @classmethod
    def unique_table_name(cls):
        cls.table_name_indx += 1
        return "siuba_{0:03d}".format(cls.table_name_indx)

    def load_df(self, df = None, **kwargs):
        df = super().load_df(df, **kwargs)

        table_name = self.unique_table_name()

        return copy_to_sql(df, table_name, self.engine)

    def load_cached_df(self, df):
        import hashlib
        from pandas import util
        hash_arr = util.hash_pandas_object(df, index=True).values
        hashed = hashlib.sha256(hash_arr).hexdigest()

        if hashed in self.cache:
            return self.cache[hashed]
        
        res = self.cache[hashed] = self.load_df(df)

        return res

class CloudBackend(SqlBackend):
    @classmethod
    def unique_table_name(cls):
        return "siuba_{}".format(uuid.uuid4())

class BigqueryBackend(CloudBackend):
    def load_df(self, df = None, **kwargs):
        df = super().load_df(df, **kwargs)

        # since we are using uuids, set table to expire after 1 day, so we can
        # easily inspect the tables, but also ensure cleanup
        self.engine.execute("""
            ALTER TABLE `{table_name}` 
            SET OPTIONS (
              expiration_timestamp=TIMESTAMP_ADD(CURRENT_TIMESTAMP(), INTERVAL 1 DAY)
            )
            """.format(table_name=df.tbl.name))
        
        return df


def robust_multiple_sort(df, by):
    """Sort a DataFrame on multiple columns, slower but more reliable than df.sort_values

    Note: pandas errors when you sort by multiple columns, and one has unhashable objects.
          however, it can "sort" a single column with unhashable objects.

          e.g. df.sort_values(by = ['a', 'b']) may cause an error

    This implementation chains sort_values on single columns. In this case,
    pandas sorts a list based on its first entry ¯\_(ツ)_/¯.
    """

    from functools import reduce

    out = reduce(lambda data, col: data.sort_values(col), by, df)

    return out.reset_index(drop = True)

def assert_frame_sort_equal(a, b, *args, **kwargs):
    """Tests that DataFrames are equal, even if rows are in different order"""
    df_a = ungroup(a)
    df_b = ungroup(b)
    sorted_a = robust_multiple_sort(df_a, list(df_a.columns)).reset_index(drop = True)
    sorted_b = robust_multiple_sort(df_b, list(df_b.columns)).reset_index(drop = True)

    assert_frame_equal(sorted_a, sorted_b, **kwargs)

def assert_equal_query(tbl, lazy_query, target, **kwargs):
    from pandas.core.groupby import DataFrameGroupBy

    out = collect(lazy_query(tbl))

    if isinstance(tbl, LazyTbl) and _is_dialect_duckdb(tbl.source):
        # TODO: find a nice way to remove duckdb specific code from here
        # duckdb does not use pandas.DataFrame.to_sql method, which coerces
        # everything to 64 bit. So we need to coerce any results it returns
        # as 32 bit to 64 bit, to match to_sql.
        int_cols = out.select_dtypes('int32').columns
        flt_cols = out.select_dtypes('float32').columns
        out[int_cols] = out[int_cols].astype('int64')
        out[flt_cols] = out[flt_cols].astype('float64')

    if isinstance(tbl, (pd.DataFrame, DataFrameGroupBy)):
        df_a = ungroup(out).reset_index(drop = True)
        df_b = ungroup(target).reset_index(drop = True)
        assert_frame_equal(df_a, df_b, **kwargs)
    else:
        assert_frame_sort_equal(out, target, **kwargs)


#PREFIX_TO_TYPE = {
#        # for datetime, need to convert to pandas datetime column
#        #"dt": types.DateTime,
#        "int": types.Integer,
#        "float": types.Float,
#        "str": types.String,
#        }

#def auto_types(df):
#    dtype = {}
#    for k in df.columns:
#        pref, *_ = k.split('_')
#        if pref in PREFIX_TO_TYPE:
#            dtype[k] = PREFIX_TO_TYPE[pref]
#    return dtype


def copy_to_sql(df, name, engine):
    if isinstance(engine, str):
        engine = sqla.create_engine(engine)

    bool_cols = [k for k, v in df.items() if v.dtype.kind == "b"]
    columns = [sqla.Column(name, sqla.types.Boolean) for name in bool_cols]


    # Create table in database ------------------------------------------------

    # TODO: clean up dialect specific work (if it grows out of control)
    if engine.dialect.name == "bigquery" and df.isna().any().any():
        # TODO: to_gbq makes all datetimes UTC, so does not round trip well, 
        # but is able to handle None -> NULL....
        project_id = engine.url.host
        qual_name = f"{engine.url.database}.{name}"
        df.to_gbq(qual_name, project_id, if_exists="replace") 
    else:
        df.to_sql(name, engine, index = False, if_exists = "replace", method="multi")


    # LazyTbl creation --------------------------------------------------------

    if engine.dialect.name == "snowflake":
        # snowflake reflection is *incredibly* slow. it could take almost 10 seconds
        # to reflect column names / types. to work around this, we tell it the 
        # column names in advance
        autoload_with = None
        columns = [sqla.Column(name) for name in df]
    else:
        autoload_with = engine

    # manually create table, so we can be explicit about boolean columns.
    # this is necessary because MySQL reflection reports them as TinyInts,
    # which mostly works, but returns ints from the query
    table = sqla.Table(
            name,
            sqla.MetaData(),
            *columns,
            autoload_with = autoload_with
            )
    
    return LazyTbl(engine, table)


from functools import wraps
import pytest
        
def backend_notimpl(*names):
    def outer(f):
        @wraps(f)
        def wrapper(backend, *args, **kwargs):
            if any(map(backend.matches_test_qualifier, names)):
                with pytest.raises((NotImplementedError, FunctionLookupError)):
                    f(backend, *args, **kwargs)
                pytest.xfail("Not implemented!")
            else:
                return f(backend, *args, **kwargs)
        return wrapper
    return outer

def backend_sql(msg):
    # allow decorating without an extra call
    if callable(msg):
        return backend_sql(None)(msg)

    def outer(f):
        @wraps(f)
        def wrapper(backend, *args, **kwargs):
            if not isinstance(backend, SqlBackend):
                pytest.skip(msg)
            else:
                return f(backend, *args, **kwargs)
        return wrapper
    return outer

def backend_pandas(msg):
    # allow decorating without an extra call
    if callable(msg):
        return backend_pandas(None)(msg)

    def outer(f):
        @wraps(f)
        def wrapper(backend, *args, **kwargs):
            if not isinstance(backend, PandasBackend):
                pytest.skip(msg)
            else:
                return f(backend, *args, **kwargs)
        return wrapper
    return outer


# Versions --------------------------------------------------------------------

re_version=r"(?P<major>\d+)\.(?P<minor>\d+).(?P<patch>\d+)"  


sqla_version=tuple(map(int, re.match(re_version, sqla.__version__).groups()))
pd_version=tuple(map(int, re.match(re_version, pd.__version__).groups()))




# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import (
        SqlColumn, SqlColumnAgg, extend_base,
        SqlTranslator,
        sql_not_impl,
        win_cumul,
        win_agg,
        annotate,
        wrap_annotate
        )
from ..utils import SQLA_VERSION

from .base import base_nowin
#from .postgresql import PostgresqlColumn as SqlColumn, PostgresqlColumnAgg as SqlColumnAgg
from . import _dt_generics as _dt

import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql
from sqlalchemy.sql import func as fn

# Custom dispatching in call trees ============================================

# Note that aggs do not inherit SqlColumnAgg, since we disable aggregate functions
# for sqlite. Could add them in, as recent versions support a wide range of aggs.
class SqliteColumn(SqlColumn): pass
class SqliteColumnAgg(SqlColumnAgg, SqliteColumn): pass


# Translations ================================================================

# fix some annotations --------------------------------------------------------

# Note this is taken from the postgres dialect, but it seems that there are 2 key points
# compared to postgresql, which always returns a float
# * sqlite date parts are returned as floats
# * sqlite time parts are returned as integers
def returns_float(func_names):
    # TODO: MC-NOTE - shift all translations to directly register
    # TODO: MC-NOTE - make an AliasAnnotated class or something, that signals
    #                 it is using another method, but w/ an updated annotation.
    from siuba.ops import ALL_OPS
    
    for name in func_names:
        generic = ALL_OPS[name]
        f_concrete = generic.dispatch(SqlColumn)
        f_annotated = wrap_annotate(f_concrete, result_type="float")
        generic.register(SqliteColumn, f_annotated)

# detect first and last date (similar to the mysql dialect) -------------------

@annotate(return_type="float")
def sql_extract(name):
    if name == "quarter":
        # division in sqlite automatically rounds down
        # so for jan, 1 + 2 = 3, and 3 / 1 is Q1
        return lambda _, col: (fn.strftime("%m", col) + 2) / 3
    return lambda _, col: fn.extract(name, col)


@_dt.sql_is_last_day_of.register
def _sql_is_last_day_of(codata: SqliteColumn, col, period):
    valid_periods = {"month",  "year"}
    if period not in valid_periods:
        raise ValueError(f"Period must be one of {valid_periods}")

    incr = f"+1 {period}"

    target_date = fn.date(col, f'start of {period}', incr, "-1 day")
    return col == target_date


@_dt.sql_is_first_day_of.register
def _sql_is_first_day_of(codata: SqliteColumn, col, period):
    valid_periods = {"month",  "year"}
    if period not in valid_periods:
        raise ValueError(f"Period must be one of {valid_periods}")

    target_date = fn.date(col, f'start of {period}')
    return fn.date(col) == target_date


# date part of period calculations --------------------------------------------

def sql_days_in_month(_, col):
    date_last_day = fn.date(col, 'start of month', '+1 month', '-1 day')
    return fn.strftime("%d", date_last_day).cast(sa_types.Integer())


def sql_week_of_year(_, col):
    # convert sqlite week to ISO week
    # adapted from: https://stackoverflow.com/a/15511864
    iso_dow = (fn.strftime("%j", fn.date(col, "-3 days", "weekday 4")) - 1)

    if SQLA_VERSION >= (2, 0, 0):
        # in v2, regular division will cause sqlalchemy to coerce the 7 to a float
        return (iso_dow // 7) + 1

    return (iso_dow / 7) + 1


# misc ------------------------------------------------------------------------
    
@annotate(result_type = "float")
def sql_round(_, col, n):
    return sql.func.round(col, n)
    

def sql_func_truediv(_, x, y):
    return sql.cast(x, sa_types.Float()) / y


def between(_, col, x, y):
    res = col.between(x, y)

    # tell sqlalchemy the result is a boolean. this causes it to be correctly
    # converted from an integer to bool when the results are collected.
    # note that this is consistent with what col == col returns
    res.type = sa_types.Boolean()
    return res

def sql_str_capitalize(_, col):
    # capitalize first letter, then concatenate with lowercased rest
    first_upper = fn.upper(fn.substr(col, 1, 1))
    rest_lower = fn.lower(fn.substr(col, 2))
    return first_upper.concat(rest_lower)

extend_base(
        SqliteColumn,

        between = between,
        clip = sql_not_impl("sqlite does not have a least or greatest function."),

        div = sql_func_truediv,
        divide = sql_func_truediv,
        rdiv = lambda _, x,y: sql_func_truediv(_, y, x),

        __truediv__ = sql_func_truediv,
        truediv = sql_func_truediv,
        __rtruediv__ = lambda _, x, y: sql_func_truediv(_, y, x),

        round = sql_round,
        __round__ = sql_round,

        **{
            "str.title": sql_not_impl("TODO"),
            "str.capitalize": sql_str_capitalize,
        },
        
        **{
            "dt.quarter": sql_extract("quarter"),
            "dt.is_quarter_start": sql_not_impl("TODO"),
            "dt.is_quarter_end": sql_not_impl("TODO"),
            "dt.days_in_month": sql_days_in_month,
            "dt.daysinmonth": sql_days_in_month,
            "dt.week": sql_week_of_year,
            "dt.weekofyear": sql_week_of_year,

        }
)

returns_float([
    "dt.dayofweek", 
    "dt.weekday",
])


extend_base(
        SqliteColumn,
        # TODO: should check sqlite version, since < 3.25 can't use windows
        cumsum = win_cumul("sum"),

        quantile = sql_not_impl("sqlite does not support ordered set aggregates"),
        sum = win_agg("sum"),
        )

extend_base(
        SqliteColumnAgg,
        quantile = sql_not_impl("sqlite does not support ordered set aggregates"),
        )


translator = SqlTranslator.from_mappings(
        SqliteColumn, SqliteColumnAgg
        )



import pytest
from .helpers import (
    assert_equal_query,
    PandasBackend,
    SqlBackend,
    CloudBackend,
    BigqueryBackend,
    data_frame
)

params_backend = [
    pytest.param(lambda: SqlBackend("postgresql"), id = "postgresql", marks=pytest.mark.postgresql),
    pytest.param(lambda: SqlBackend("mysql"), id = "mysql", marks=pytest.mark.mysql),
    pytest.param(lambda: SqlBackend("sqlite"), id = "sqlite", marks=pytest.mark.sqlite),
    pytest.param(lambda: SqlBackend("duckdb"), id = "duckdb", marks=pytest.mark.duckdb),
    pytest.param(lambda: BigqueryBackend("bigquery"), id = "bigquery", marks=pytest.mark.bigquery),
    pytest.param(lambda: CloudBackend("snowflake"), id = "snowflake", marks=pytest.mark.snowflake),
    pytest.param(lambda: PandasBackend("pandas"), id = "pandas", marks=pytest.mark.pandas)
    ]

@pytest.fixture(params = params_backend, scope = "session")
def backend(request):
    return request.param()

@pytest.fixture
def skip_backend(request, backend):
    if request.node.get_closest_marker('skip_backend'):
        mark_args = request.node.get_closest_marker('skip_backend').args
        if backend.name in mark_args:
            pytest.skip('skipped on backend: {}'.format(backend.name)) 

@pytest.fixture
def xfail_backend(request, backend):
    if request.node.get_closest_marker('xfail_backend'):
        mark_args = request.node.get_closest_marker('xfail_backend').args
        if backend.name in mark_args:
            pytest.xfail()




import warnings

from sqlalchemy import sql

from siuba.dply.verbs import case_when, if_else
from siuba.siu import Call

from ..utils import _sql_case
from ..backend import LazyTbl
from ..translate import ColumnCollection


@case_when.register(ColumnCollection)
def _case_when(__data, cases):
    # TODO: will need listener to enter case statements, to handle when they use windows
    if isinstance(cases, Call):
        cases = cases(__data)

    whens = []
    case_items = list(cases.items())
    n_items = len(case_items)

    else_val = None
    for ii, (expr, val) in enumerate(case_items):
        # handle where val is a column expr
        if callable(val):
            val = val(__data)

        # handle when expressions
        #if ii+1 == n_items and expr is True:
        #    else_val = val
        if expr is True:
            # note: only sqlalchemy v1.3 requires wrapping in literal
            whens.append((sql.literal(expr), val))
        elif callable(expr):
            whens.append((expr(__data), val))
        else:
            whens.append((expr, val))

    return _sql_case(whens, else_ = else_val)


@case_when.register(LazyTbl)
def _case_when(__data, cases):
    raise NotImplementedError(
        "`case_when()` must be used inside a verb like `mutate()`, when using a "
        "SQL backend."
    )
        

# if_else ---------------------------------------------------------------------

@if_else.register(sql.elements.ColumnElement)
def _if_else(cond, true_vals, false_vals):
    whens = [(cond, true_vals)]
    return _sql_case(whens, else_ = false_vals)



from .pivot_long import pivot_longer, build_longer_spec, pivot_longer_spec
from .pivot_wide import pivot_wider, build_wider_spec, pivot_wider_spec

from . import sql_pivot_long
from . import sql_pivot_wide



"""Functions for creating fast versions of methods like .mean().

Examples:
    .. code-block:: python    

       from siuba.data import mtcars
       g_cyl = mtcars.groupby("cyl")

       avg_hp_raw = g_cyl.hp.mean()

       # imagine that avg_hp was not SeriesGroupBy, but a GroupByAgg object
       avg_hp = GroupByAgg.from_result(avg_hp_raw, g_cyl.hp)

       f_mean = method_agg_op("mean", is_property = False, accessor = None)
       f_mean(g_cyl.hp)        # returns GroupByAgg

       f_add = method_el_op2("__add__", is_property = False, accessor = None)
       f_add(g_cyl.hp, g_cyl.hp)
       f_add(f_mean(g_cyl.hp), g_cyl.hp)

       # property methods ----
       f_is_unique = method_el_op("is_unique", is_property = True, None)
       f_is_unique(g_cyl.hp)

       # accessor methods ----
       import pandas as pd
       gdf = pd.DataFrame({'g': ['a', 'b'], 'x': ['AA', 'BB']}).groupby('g')

       f_str_lower = method_el_op("lower", False, "str")
       f_str_lower(gdf.x)

"""

from siuba.siu import FunctionLookupBound
from .groupby import GroupByAgg, SeriesGroupBy, broadcast_group_elements, regroup

import pandas as pd


# utilities -------------------------------------------------------------------

def _validate_data_args(x, **kwargs):
    # Sanity checks that only SeriesGroupBy passed for ops
    # could be removed once this sees more use
    if not isinstance(x, SeriesGroupBy):
        raise TypeError("First data argument must be a grouped Series object")

    for name, other_data in kwargs.items():
        if isinstance(other_data, pd.Series):
            raise TypeError("{} may not be a Series.".format(name))


def _apply_grouped_method(ser, name, is_property, accessor, args, kwargs):
    if accessor:
        method = getattr(getattr(ser, accessor), name)
    else:
        method = getattr(ser, name)

    res = method(*args, **kwargs) if not is_property else method

    return res


# Translations ----------------------------------------------------------------

def not_implemented(name, *args, **kwargs):
    return FunctionLookupBound(name)


def method_agg_op(name, is_property, accessor):
    def f(__ser: SeriesGroupBy, *args, **kwargs) -> GroupByAgg:
        _validate_data_args(__ser)
        res = _apply_grouped_method(__ser, name, is_property, accessor, args, kwargs)

        return GroupByAgg.from_result(res, __ser)

    f.__name__ = f.__qualname__ = name
    return f


def method_el_op(name, is_property, accessor):
    def f(__ser: SeriesGroupBy, *args, **kwargs) -> SeriesGroupBy:
        _validate_data_args(__ser)
        res = _apply_grouped_method(__ser.obj, name, is_property, accessor, args, kwargs)

        return regroup(__ser, res)

    f.__name__ = f.__qualname__ = name
    return f


def method_el_op2(name, is_property, accessor):
    def f(x: SeriesGroupBy, y: SeriesGroupBy, *args, **kwargs) -> SeriesGroupBy:
        _validate_data_args(x, y = y)
        left, right, ref_groupby = broadcast_group_elements(x, y)
        
        op_function = getattr(left, name)

        res = op_function(right, *args, **kwargs)
        return regroup(ref_groupby, res)

    f.__name__ = f.__qualname__ = name
    return f


def method_win_op(name, is_property, accessor):
    def f(__ser: SeriesGroupBy, *args, **kwargs) -> SeriesGroupBy:
        _validate_data_args(__ser)
        res = _apply_grouped_method(__ser, name, is_property, accessor, args, kwargs)

        return regroup(__ser, res)

    f.__name__ = f.__qualname__ = name
    return f

def method_win_op_agg_result(name, is_property, accessor):
    def f(__ser, *args, **kwargs):
        _validate_data_args(__ser)
        res = _apply_grouped_method(__ser, name, is_property, accessor, args, kwargs)

        return GroupByAgg.from_result(res, __ser)

    f.__name__ = f.__qualname__ = name
    return f


def method_agg_singleton(name, is_property, accessor):
    def f(__ser: SeriesGroupBy, *args, **kwargs) -> SeriesGroupBy:
        _validate_data_args(__ser)
        if accessor is not None:
            op_function = getattr(getattr(__ser.obj, accessor, __ser.obj), name)
        else:
            op_function = getattr(__ser.obj, name)

        # cast singleton result to be GroupByAgg, as if we did an aggregation
        # could create a class to for grouped singletons, but seems like overkill
        # for now
        singleton = op_function if is_property else op_function()
        dtype = 'object' if singleton is None else None

        # note that when the value is None, need to explicitly make dtype object
        res = pd.Series(singleton, index = __ser.grouper.levels, dtype = dtype) 

        return GroupByAgg.from_result(res, __ser)
    return f


def forward_method(dispatcher, constructor = None, cls = SeriesGroupBy):
    op = dispatcher.operation
    kind = op.kind.title() if op.kind is not None else None
    key = (kind, op.arity)

    constructor = GROUP_METHODS[key] if constructor is None else constructor

    f_concrete = constructor(
            name = op.name,
            is_property = op.is_property,
            accessor = op.accessor
            )
    
    return dispatcher.register(cls, f_concrete)



# TODO: which methods can be pulled off GroupedDFs?
#       * all elwise in certain categories can be added (dt, str, etc..)
#       * some aggs / windows can be punted to SeriesGroupBy
#       * others require custom implementation

GROUP_METHODS = { 
        ("Elwise", 1): method_el_op,
        ("Elwise", 2): method_el_op2,
        ("Agg", 1): method_agg_op,
        ("Agg", 2): not_implemented,
        ("Window", 1): method_win_op,
        ("Window", 2): method_win_op,
        ("Singleton", 1): method_agg_singleton,
        ("Todo", 1): not_implemented,
        ("Maydo", 1): not_implemented,
        ("Wontdo", 1): not_implemented,
        ("Todo", 2): not_implemented,
        ("Maydo", 2): not_implemented,
        ("Wontdo", 2): not_implemented,
        (None, 1): not_implemented,
        (None, 2): not_implemented,
        }






from siuba.siu import symbolic_dispatch
from pandas.arrays import DatetimeArray, PeriodArray
from pandas import DatetimeIndex, PeriodIndex, Period, Timestamp, Series

from pandas import offsets
from pandas.core.dtypes.common import is_period_dtype, is_datetime64_any_dtype

from abc import ABC

try:
    from pandas.tseries.offsets import prefix_mapping
except ImportError:
    from pandas._libs.tslibs.offsets import prefix_mapping

LUBRIDATE_OFFSETS = {
        "second": "S",
        "minute": "M",
        "hour": "H",
        "day": "D",
        "week": "W",
        "month": "M",
        "bimonth": "2M",
        "quarter": "Q",
        "season": None,
        "halfyear": None,
        "year": "Y"
        }

# There's no class that clearly identifies all the Datetime op compatible classes,
# so make an ABC and register them on it
def _make_abc(name, subclasses):
    cls = type(name, (ABC,), {})
    for child in subclasses: cls.register(child)
    return cls

DatetimeType = _make_abc("DatetimeType", [DatetimeIndex, DatetimeArray, Timestamp])
PeriodType   = _make_abc("PeriodType",   [PeriodIndex, PeriodArray, Period])

def _get_offset(unit):
    cls = prefix_mapping.get(unit)
    if cls is None:
        raise ValueError("unit {} not a valid offset".format(unit))

    return cls

def _get_series_dispatcher(f, x):
    if is_period_dtype(x):
        return f.registry[PeriodType]
    elif is_datetime64_any_dtype(x):
        return f.registry[DatetimeType]

    raise TypeError("does not seem to be a period or datetime")

DOCSTRING = """\
    floor_date and ceil_date return dates rounded to nearest specified unit.

    These functions address limitations in pandas' round() method, which cannot
    round down to, e.g., the month.
    
    Parameters
    ----------
    x:
        a DatetimeIndex, PeriodIndex, or their underlying arrays or elements.
    units:
        a date or time unit for rounding (eg. "MS" rounds down or up to the start of a month)
    
    Notes
    -----
    
    For a full list of units run the following:
    
        from pandas.tseries.offsets import prefix_mapping
    
    Common time units include seconds (S), minute (T), hour (H), week (W). 
    Some date units are month end (M), month start (MS), year start (AS)
    
    
    The main feature of floor_date is that it works on things like Months, which is
    not supported in the floor method.
    
    Examples
    --------
    
    >>> import pandas as pd
    >>> a_date = "2020-02-02 02:02:02"
    >>> dti = pd.DatetimeIndex([a_date])
    >>> dti.floor("MS")
    Traceback (most recent call last):
    ...
    ValueError: <MonthBegin> is a non-fixed frequency
    
    Month start will always go to the first day of a month.
    
    >>> floor_date(dti, "MS") 
    DatetimeIndex(['2020-02-01'], dtype='datetime64[ns]', freq=None)
    
    >>> ceil_date(dti, "MS")
    DatetimeIndex(['2020-03-01'], dtype='datetime64[ns]', freq=None)
    
    On the other hand, here is month end.
    
    >>> floor_date(dti, "M") 
    DatetimeIndex(['2020-01-31'], dtype='datetime64[ns]', freq=None)
    
    >>> ceil_date(dti, "M")
    DatetimeIndex(['2020-02-29'], dtype='datetime64[ns]', freq=None)
    
    It also works on things supported by the Series.dt.floor method, like hours.
    
    >>> floor_date(dti, "H")
    DatetimeIndex(['2020-02-02 02:00:00'], dtype='datetime64[ns]', freq=None)
    
    You can also use it on other types, like a PeriodIndex
    
    >>> per = pd.PeriodIndex([a_date], freq = "S")
    >>> floor_date(per, "M")
    PeriodIndex(['2020-02'], dtype='period[M]'...
    
    """

# Create single dispatch functions --------------------------------------------

@symbolic_dispatch
def floor_date(x, unit = "S"):
    raise TypeError("floor_date not implemented for class {}".format(type(x)))


@symbolic_dispatch
def ceil_date(x, unit = "S"):
    raise TypeError("ceil_date not implemented for class {}".format(type(x)))

floor_date.__doc__ = DOCSTRING
floor_date.__wrapped__.__doc__ = DOCSTRING

ceil_date.__doc__  = DOCSTRING
ceil_date.__wrapped__.__doc__ = DOCSTRING


# Floor date ------------------------------------------------------------------

@floor_date.register(DatetimeType)
def _(x, unit = "S"):
    cls_offset = _get_offset(unit)

    if issubclass(cls_offset, offsets.Tick):
        return x.floor(unit)

    # note: x - 0*offset shifts anchor forward for some reason, so we
    # add then subtract to ensure it doesn't change anchor points
    offset = cls_offset(n = 1)
    return x.normalize() + offset - offset          


@floor_date.register(PeriodType)
def _(x, unit = "S"):
    return x.asfreq(unit, how = "start")


@floor_date.register(Series)
def _(x, *args, **kwargs):
    # dispatch to either period or datetime version
    f = _get_series_dispatcher(floor_date, x)
    return f(x.dt, *args, **kwargs)


# Ceil date -------------------------------------------------------------------
 
@ceil_date.register(DatetimeType)
def _(x, unit = "S"):
    cls_offset = _get_offset(unit)

    if issubclass(cls_offset, offsets.Tick):
        return x.ceil(unit)

    offset = cls_offset(n = 0)
    # the 0 ensures it will not rollforward an anchor point
    return x.normalize() + offset


@ceil_date.register(PeriodType)
def _(x, unit = "S"):
    raise NotImplementedError(
            "It is not possible to use ceil_date on a Period. "
            "Try converting to a DatetimeIndex."
            )


@ceil_date.register(Series)
def _(x, *args, **kwargs):
    f = _get_series_dispatcher(ceil_date, x)
    return f(x.dt, *args, **kwargs)





import warnings

from sqlalchemy.sql.elements import ClauseElement 
from sqlalchemy import sql

from ..translate import (
        SqlColumn, SqlColumnAgg,
        win_cumul, AggOver, CumlOver, RankOver, warn_arg_default, win_absent
        )

from ..dialects.sqlite import SqliteColumn
from ..dialects.mysql import MysqlColumn
from ..dialects.bigquery import BigqueryColumn
from ..dialects.duckdb import DuckdbColumn

from siuba.dply.vector import (
        #cumall, cumany, cummean,
        desc,
        dense_rank, percent_rank, min_rank, cume_dist,
        row_number,
        #ntile,
        between,
        coalesce,
        lead, lag,
        n,
        n_distinct,
        na_if,
        #near,
        nth, first, last
        )


from ..translate import SiubaSqlRuntimeWarning
warnings.simplefilter('once', SiubaSqlRuntimeWarning)

# desc ------------------------------------------------------------------------

@desc.register
def _desc_sql(codata: SqlColumn, x: ClauseElement) -> ClauseElement:
    """
    Example:
        >>> print(desc(SqlColumn(), sql.column('a')))
        a DESC
    """
    return x.desc()


# ranking functions -----------------------------------------------------------
# note: here we don't use the decorator syntax, but maybe we should for
#       consistency
# TODO: remove repetition in rank definitions

def _sql_rank_over(rank_func, col, partition, nulls_last):
    # partitioning ensures aggregates that use total length are correct,
    # e.g. percent rank, cume_dist and friends, by separating NULLs into their 
    # own partition
    over_clause = RankOver(
            rank_func(),
            order_by = col if not nulls_last else col.nullslast(),
            partition_by = col.isnot(None) if partition else None
            )

    return sql.case({col.isnot(None): over_clause})

def _sql_rank(func_name, partition = False, nulls_last = False):
    # partition controls whether to make partition by NOT NULL

    rank_func = getattr(sql.func, func_name)

    def f(_, col, na_option = None) -> RankOver:
        if na_option == "keep":
            return _sql_rank_over(rank_func, col, partition, nulls_last)

        warn_arg_default(func_name, 'na_option', None, "keep")

        return RankOver(rank_func(), order_by = col)

    return f


dense_rank  .register(SqlColumn, _sql_rank("dense_rank"))
percent_rank.register(SqlColumn, _sql_rank("percent_rank"))
cume_dist   .register(SqlColumn, _sql_rank("cume_dist", partition = True))
min_rank    .register(SqlColumn, _sql_rank("rank", partition = True))


dense_rank  .register(SqliteColumn, _sql_rank("dense_rank", nulls_last=True))
percent_rank.register(SqliteColumn, _sql_rank("percent_rank", nulls_last=True))
cume_dist   .register(SqliteColumn, _sql_rank("cume_dist", partition = True))
min_rank    .register(SqliteColumn, _sql_rank("rank", nulls_last=True))

# partition everything, since MySQL puts NULLs first
# see: https://stackoverflow.com/q/1498648/1144523
dense_rank  .register(MysqlColumn, _sql_rank("dense_rank", partition = True))
percent_rank.register(MysqlColumn, _sql_rank("percent_rank", partition = True))
cume_dist   .register(MysqlColumn, _sql_rank("cume_dist", partition = True))
min_rank    .register(MysqlColumn, _sql_rank("rank", partition = True))

dense_rank  .register(BigqueryColumn, _sql_rank("dense_rank", nulls_last = True))
percent_rank.register(BigqueryColumn, _sql_rank("percent_rank", nulls_last = True))

dense_rank  .register(DuckdbColumn, _sql_rank("dense_rank", nulls_last = True))

# row_number ------------------------------------------------------------------

@row_number.register
def _row_number_sql(codata: SqlColumn, col: ClauseElement) -> CumlOver:
    """
    Example:
        >>> print(row_number(SqlColumn(), sql.column('a')))
        row_number() OVER ()
        
    """
    return CumlOver(sql.func.row_number())

# between ---------------------------------------------------------------------

@between.register
def _between_sql(codata: SqlColumn, x, left, right, default = None) -> ClauseElement:
    """
    Example:
        >>> print(between(SqlColumn(), sql.column('a'), 1, 2))
        a BETWEEN :a_1 AND :a_2

        >>> print(between(SqlColumn(), sql.column('a'), 1, 2, default = False))
        coalesce(a BETWEEN :a_1 AND :a_2, :coalesce_1)

    """
    
    if default is not False:
        # TODO: warn
        pass

    if default is None:
        return x.between(left, right)

    return sql.functions.coalesce(x.between(left, right), default)

# coalesce --------------------------------------------------------------------

@coalesce.register
def _coalesce_sql(codata: SqlColumn, x, *args) -> ClauseElement:
    """
    Example:
        >>> print(coalesce(SqlColumn(), sql.column('a'), sql.column('b')))
        coalesce(a, b)

        >>> print(coalesce(SqlColumn(), 1, sql.column('a')))
        coalesce(:coalesce_1, a)
    """
    return sql.functions.coalesce(x, *args)


# lead and lag ----------------------------------------------------------------

@lead.register
def _lead_sql(codata: SqlColumn, x, n = 1, default = None) -> ClauseElement:
    """
    Example:
        >>> print(lead(SqlColumn(), sql.column('a'), 2, 99))
        lead(a, :lead_1, :lead_2) OVER ()
    """
    
    f = win_cumul("lead", rows=None)
    return f(codata, x, n, default)

@lag.register
def _lag_sql(codata: SqlColumn, x, n = 1, default = None) -> ClauseElement:
    """
    Example:
        >>> print(lag(SqlColumn(), sql.column('a'), 2, 99))
        lag(a, :lag_1, :lag_2) OVER ()
    """
    f = win_cumul("lag", rows=None)
    return f(codata, x, n , default)


# n ---------------------------------------------------------------------------

@n.register
def _n_sql(codata: SqlColumn, x) -> ClauseElement:
    """
    Example:
        >>> print(n(SqlColumn(), sql.column('a')))
        count(*) OVER ()
    """
    return AggOver(sql.func.count())

# TODO: MC-Note - fix

@n.register
def _n_sql_agg(codata: SqlColumnAgg, x) -> ClauseElement:
    """
    Example:
        >>>
        >> from siuba.sql.translate import SqlColumnAgg
        >> print(n(SqlColumnAgg(), None))
        count(*)
    """

    return sql.func.count()



# n_distinct ------------------------------------------------------------------

@n_distinct.register
def _n_distinct_sql(codata: SqlColumn, x: ClauseElement) -> ClauseElement:
    """
    Example:
        >>> print(n_distinct(SqlColumn(), sql.column('a')) )
        count(distinct(a))
    """
    return sql.func.count(sql.func.distinct(x))


# na_if -----------------------------------------------------------------------

@na_if.register
def _na_if_sql(codata: SqlColumn, x, y) -> ClauseElement:
    """
    Example:
        >>> print(na_if(SqlColumn(), sql.column('x'), 2))
        nullif(x, :nullif_1)
    """
    return sql.func.nullif(x, y)


# nth, first, last ------------------------------------------------------------
# note: first and last wrap around nth, so are not dispatchers.
#       this may need to change this in the future, since this means they won't
#       show their own name, when you print, e.g. first(_.x)

@nth.register
def _nth_sql(codata: SqlColumn, x, n, order_by = None, default = None) -> ClauseElement:
    if default is not None:
        raise NotImplementedError("default argument not implemented")

    if n < 0 and order_by is None:
        raise NotImplementedError(
                "must explicitly pass order_by when using last or nth with "
                "n < 0 in SQL."
                )

    if n < 0:
        # e.g. -1 in python is 0, -2 is 1.
        n = abs(n + 1)
        order_by = order_by.desc()


    #  note the adjustment for 1-based index in SQL
    return CumlOver(
            sql.func.nth_value(x, n + 1),
            order_by = order_by,
            rows = (None, None)
            )


@nth.register
def _nth_sql_agg(codata: SqlColumnAgg, x, n, order_by = None, default = None) -> ClauseElement:
    raise NotImplementedError("nth, first, and last not available in summarize")

